import time
from metadrive.envs.marl_envs.multi_agent_metadrive import MULTI_AGENT_METADRIVE_DEFAULT_CONFIG
MULTI_AGENT_METADRIVE_DEFAULT_CONFIG["force_seed_spawn_manager"] = True
import numpy as np
from gymnasium.spaces import Box, Dict

from metadrive.envs.marl_envs.marl_parking_lot import MultiAgentParkingLotEnv
from metadrive.utils import distance_greater, norm


def _check_spaces_before_reset(env):
    a = set(env.config["target_vehicle_configs"].keys())
    b = set(env.observation_space.spaces.keys())
    c = set(env.action_space.spaces.keys())
    assert a == b == c
    _check_space(env)


def _check_spaces_after_reset(env, obs=None):
    a = set(env.config["target_vehicle_configs"].keys())
    b = set(env.observation_space.spaces.keys())
    assert a == b
    _check_shape(env)

    if obs:
        assert isinstance(obs, dict)
        assert set(obs.keys()) == a


def _check_shape(env):
    b = set(env.observation_space.spaces.keys())
    c = set(env.action_space.spaces.keys())
    d = set(env.vehicles.keys())
    e = set(env.engine.agents.keys())
    f = set([k for k in env.observation_space.spaces.keys() if not env.dones[k]])
    assert d == e == f, (b, c, d, e, f)
    assert c.issuperset(d)
    _check_space(env)


def _check_space(env):
    assert isinstance(env.action_space, Dict)
    assert isinstance(env.observation_space, Dict)
    o_shape = None
    for k, s in env.observation_space.spaces.items():
        assert isinstance(s, Box)
        if o_shape is None:
            o_shape = s.shape
        assert s.shape == o_shape
    a_shape = None
    for k, s in env.action_space.spaces.items():
        assert isinstance(s, Box)
        if a_shape is None:
            a_shape = s.shape
        assert s.shape == a_shape


def _act(env, action):
    assert env.action_space.contains(action)
    obs, reward, terminated, truncated, info = env.step(action)
    _check_shape(env)
    if not terminated["__all__"]:
        assert len(env.vehicles) > 0
    if not (set(obs.keys()) == set(reward.keys()) == set(env.observation_space.spaces.keys())):
        raise ValueError
    assert env.observation_space.contains(obs)
    assert isinstance(reward, dict)
    assert isinstance(info, dict)
    assert isinstance(terminated, dict)
    assert isinstance(truncated, dict)

    return obs, reward, terminated, truncated, info


def test_ma_parking_lot_env():
    for env in [MultiAgentParkingLotEnv({"delay_done": 0, "num_agents": 1,
                                         "vehicle_config": {"lidar": {"num_others": 8}}}),
                MultiAgentParkingLotEnv({"num_agents": 1, "delay_done": 0,
                                         "vehicle_config": {"lidar": {"num_others": 0}}}),
                MultiAgentParkingLotEnv({"num_agents": 4, "delay_done": 0,
                                         "vehicle_config": {"lidar": {"num_others": 8}}}),
                MultiAgentParkingLotEnv({"num_agents": 4, "delay_done": 0,
                                         "vehicle_config": {"lidar": {"num_others": 0}}}),
                MultiAgentParkingLotEnv({"num_agents": 8, "delay_done": 0,
                                         "vehicle_config": {"lidar": {"num_others": 0}}})]:
        try:
            _check_spaces_before_reset(env)
            obs, _ = env.reset()
            _check_spaces_after_reset(env, obs)
            assert env.observation_space.contains(obs)
            for step in range(100):
                act = {k: [1, 1] for k in env.vehicles.keys()}
                o, r, tm, tc, i = _act(env, act)
                if step == 0:
                    assert not any(tm.values())
                    assert not any(tc.values())

        finally:
            env.close()


def test_ma_parking_lot_horizon():
    # test horizon
    for _ in range(3):  # This function is really easy to break, repeat multiple times!
        env = MultiAgentParkingLotEnv(
            {
                "horizon": 100,
                "num_agents": 4,
                "vehicle_config": {
                    "lidar": {
                        "num_others": 2
                    }
                },
                "out_of_road_penalty": 777,
                "out_of_road_cost": 778,
                "crash_done": False
            }
        )
        try:
            _check_spaces_before_reset(env)
            obs, _ = env.reset()
            _check_spaces_after_reset(env, obs)
            assert env.observation_space.contains(obs)
            last_keys = set(env.vehicles.keys())
            for step in range(1, 1000):
                act = {k: [1, 1] for k in env.vehicles.keys()}
                o, r, tm, tc, i = _act(env, act)
                new_keys = set(env.vehicles.keys())
                if step == 0:
                    assert not any(tm.values())
                    assert not any(tc.values())
                if any(tm.values()):
                    assert len(last_keys) <= 4  # num of agents
                    assert len(new_keys) <= 4  # num of agents
                    for k in new_keys.difference(last_keys):
                        assert k in o
                        assert k in tm
                    # print("Step {}, Done: {}".format(step, d))

                for kkk, rrr in r.items():
                    if rrr == -777:
                        assert tm[kkk]
                        assert i[kkk]["cost"] == 778
                        assert i[kkk]["out_of_road"]

                for kkk, iii in i.items():
                    if "out_of_road" in iii and (iii["out_of_road"] or iii["cost"] == 778):
                        assert tm[kkk]
                        assert i[kkk]["cost"] == 778
                        assert i[kkk]["out_of_road"]
                        #assert r[kkk] == -777

                if tm["__all__"]:
                    break
                last_keys = new_keys
        finally:
            env.close()


def test_ma_parking_lot_reset():
    env = MultiAgentParkingLotEnv({"horizon": 50, "num_agents": 11})
    try:
        _check_spaces_before_reset(env)
        obs, _ = env.reset()
        _check_spaces_after_reset(env, obs)
        assert env.observation_space.contains(obs)
        for step in range(1000):
            act = {k: [1, 1] for k in env.vehicles.keys()}
            o, r, tm, tc, i = _act(env, act)
            if step == 0:
                assert not any(tm.values())
                assert not any(tc.values())
            if tm["__all__"]:
                obs, _ = env.reset()
                assert env.observation_space.contains(obs)

                _check_spaces_after_reset(env, obs)
                assert set(env.observation_space.spaces.keys()) == set(env.action_space.spaces.keys()) == \
                       set(env.observations.keys()) == set(obs.keys()) == \
                       set(env.config["target_vehicle_configs"].keys())

                break
    finally:
        env.close()

    # Put vehicles to destination and then reset. This might cause error if agent is assigned destination BEFORE reset.
    env = MultiAgentParkingLotEnv({"horizon": 100, "num_agents": 11, "success_reward": 777})
    try:
        _check_spaces_before_reset(env)
        success_count = 0
        agent_count = 0
        obs, _ = env.reset()
        _check_spaces_after_reset(env, obs)
        assert env.observation_space.contains(obs)

        for num_reset in range(5):
            for step in range(1000):

                # for _ in range(2):
                #     act = {k: [1, 1] for k in env.vehicles.keys()}
                #     o, r, tm, tc, i = _act(env, act)

                # Force vehicle to success!
                for v_id, v in env.vehicles.items():
                    loc = v.navigation.final_lane.end
                    # vehicle will stack together to explode!
                    v.set_position(loc, height=int(v_id[5:]) * 2)
                    v.set_position(loc)
                    pos = v.position
                    np.testing.assert_almost_equal(pos, loc, decimal=3)
                    new_loc = v.navigation.final_lane.end
                    long, lat = v.navigation.final_lane.local_coordinates(v.position)
                    flag1 = (v.navigation.final_lane.length - 5 < long < v.navigation.final_lane.length + 5)
                    flag2 = (
                        v.navigation.get_current_lane_width() / 2 >= lat >=
                        (0.5 - v.navigation.get_current_lane_num()) * v.navigation.get_current_lane_width()
                    )
                    # if not env._is_arrive_destination(v):
                    # print('sss')
                    assert env._is_arrive_destination(v)

                act = {k: [0, 0] for k in env.vehicles.keys()}
                o, r, tm, tc, i = _act(env, act)

                for v in env.vehicles.values():
                    assert len(v.navigation.checkpoints) > 2

                for kkk, iii in i.items():
                    if "arrive_dest" in iii and iii["arrive_dest"]:
                        # # print("{} success!".format(kkk))
                        success_count += 1

                for kkk, ddd in tm.items():
                    if ddd and kkk != "__all__":
                        assert i[kkk]["arrive_dest"]
                        agent_count += 1

                for kkk, rrr in r.items():
                    if tm[kkk]:
                        assert rrr == 777

                if tm["__all__"]:
                    # print("Finish {} agents. Success {} agents.".format(agent_count, success_count))
                    o, _ = env.reset()
                    assert env.observation_space.contains(o)
                    _check_spaces_after_reset(env, o)
                    break
    finally:
        env.close()


def test_ma_parking_lot_close_spawn():
    def _no_close_spawn(vehicles):
        vehicles = list(vehicles.values())
        for c1, v1 in enumerate(vehicles):
            for c2 in range(c1 + 1, len(vehicles)):
                v2 = vehicles[c2]
                dis = norm(v1.position[0] - v2.position[0], v1.position[1] - v2.position[1])
                assert distance_greater(v1.position, v2.position, length=2.2)

    MultiAgentParkingLotEnv._DEBUG_RANDOM_SEED = 1
    env = MultiAgentParkingLotEnv({
        # "use_render": True,
        "horizon": 50,
        "num_agents": 11,
    })
    env.seed(100)
    try:
        _check_spaces_before_reset(env)
        for num_r in range(10):
            obs, _ = env.reset()
            _check_spaces_after_reset(env)
            for _ in range(10):
                o, r, tm, tc, i = env.step({k: [0, 0] for k in env.vehicles.keys()})
                assert not any(tm.values())
                assert not any(tc.values())
            _no_close_spawn(env.vehicles)
            # print('Finish {} resets.'.format(num_r))
    finally:
        env.close()
        MultiAgentParkingLotEnv._DEBUG_RANDOM_SEED = None


def test_ma_parking_lot_reward_done_alignment():
    # out of road
    env = MultiAgentParkingLotEnv({"horizon": 200, "num_agents": 11, "out_of_road_penalty": 777, "crash_done": False})
    try:
        _check_spaces_before_reset(env)
        obs, _ = env.reset(seed=0)
        _check_spaces_after_reset(env, obs)
        assert env.observation_space.contains(obs)
        out_num = 0
        for action in [-1, 1]:
            for step in range(5000):
                act = {k: [action, 1] for k in env.vehicles.keys()}
                o, r, tm, tc, i = _act(env, act)
                for kkk, ddd in tm.items():
                    if ddd and kkk != "__all__":
                        #assert r[kkk] == -777
                        assert i[kkk]["out_of_road"] or i[kkk]["max_step"]
                        if i[kkk]["out_of_road"]:
                            out_num += 1
                        # # print('{} done passed!'.format(kkk))
                for kkk, rrr in r.items():
                    if rrr == -777:
                        assert tm[kkk]
                        assert i[kkk]["out_of_road"]
                        # # print('{} reward passed!'.format(kkk))
                if tm["__all__"]:
                    env.reset(seed=0)
                    break
        assert out_num > 10
    finally:
        env.close()

    # crash
    env = MultiAgentParkingLotEnv(
        {
            "horizon": 100,
            "num_agents": 11,
            "crash_vehicle_penalty": 1.7777,
            "crash_done": True,
            "delay_done": 0,

            # "use_render": True,
            #
            "top_down_camera_initial_z": 160
        }
    )
    # Force the seed here so that the agent1 and agent2 are in same heading! Otherwise they might be in vertical
    # heading and cause one of the vehicle raise "out of road" error!
    env._DEBUG_RANDOM_SEED = 1
    try:
        _check_spaces_before_reset(env)
        obs, _ = env.reset()
        _check_spaces_after_reset(env, obs)
        for step in range(5):
            act = {k: [0, 0] for k in env.vehicles.keys()}
            o, r, tm, tc, i = _act(env, act)
        env.vehicles["agent0"].set_position(env.vehicles["agent1"].position, height=1.2)
        for step in range(5000):
            act = {k: [0, 0] for k in env.vehicles.keys()}
            o, r, tm, tc, i = _act(env, act)

            if not any(tm.values()):
                continue

            # assert sum(d.values()) == 2

            for kkk in ['agent0', 'agent1']:
                iii = i[kkk]
                assert iii["crash_vehicle"]
                assert iii["crash"]
                #assert r[kkk] == -1.7777
                # for kkk, ddd in d.items():
                ddd = tm[kkk]
                if ddd and kkk != "__all__":
                    #assert r[kkk] == -1.7777
                    assert i[kkk]["crash_vehicle"]
                    assert i[kkk]["crash"]
                    # # print('{} done passed!'.format(kkk))
                # for kkk, rrr in r.items():
                rrr = r[kkk]
                if rrr == -1.7777:
                    assert tm[kkk]
                    assert i[kkk]["crash_vehicle"]
                    assert i[kkk]["crash"]
                    # # print('{} reward passed!'.format(kkk))
            # assert d["__all__"]
            # if d["__all__"]:
            break
    finally:
        env._DEBUG_RANDOM_SEED = None
        env.close()

    # crash with real fixed vehicle

    # crash 2
    env = MultiAgentParkingLotEnv(
        {
            "map_config": {
                "lane_num": 1
            },
            # "use_render": True,
            #
            "horizon": 200,
            "num_agents": 7,
            "crash_vehicle_penalty": 1.7777,
            "parking_space_num": 16,
            "crash_done": False,
            "use_render": False
        }
    )
    try:
        _check_spaces_before_reset(env)
        obs, _ = env.reset()
        env.engine.spawn_manager.np_random = np.random.RandomState(0)
        obs, _ = env.reset(seed=0)
        _check_spaces_after_reset(env, obs)
        for step in range(1):
            act = {k: [0, 0] for k in env.vehicles.keys()}
            o, r, tm, tc, i = _act(env, act)

        for v_id, v in env.vehicles.items():
            if v_id != "agent0":
                v.set_static(True)
        out_num = 0
        for step in range(5000):
            act = {k: [0, 1] for k in env.vehicles.keys()}
            o, r, tm, tc, i = _act(env, act)
            for kkk, iii in i.items():
                if iii["crash"]:
                    assert iii["crash_vehicle"]
                if iii["crash_vehicle"]:
                    assert iii["crash"]
                    # #assert r[kkk] == -1.7777
            for kkk, ddd in tm.items():
                if ddd and kkk != "__all__":
                    assert i[kkk]["out_of_road"] or i[kkk]["max_step"]
                    if i[kkk]["out_of_road"]:
                        out_num += 1
                    # # print('{} done passed!'.format(kkk))
            for kkk, rrr in r.items():
                if rrr == -1.7777:
                    # assert d[kkk]
                    assert i[kkk]["crash_vehicle"]
                    assert i[kkk]["crash"]
                    # # print('{} reward passed!'.format(kkk))
            if tm["agent0"]:
                break
            if tm["__all__"]:
                break
        assert out_num > 0
    finally:
        env.close()

    # success
    env = MultiAgentParkingLotEnv(
        {
            "horizon": 100,
            "num_agents": 2,
            "success_reward": 999,
            "out_of_road_penalty": 555,
            "crash_done": True
        }
    )
    try:
        _check_spaces_before_reset(env)
        obs, _ = env.reset()
        _check_spaces_after_reset(env)
        env.vehicles["agent0"].set_position(env.vehicles["agent0"].navigation.final_lane.end)
        assert env.observation_space.contains(obs)
        for step in range(5000):
            act = {k: [0, 0] for k in env.vehicles.keys()}
            o, r, tm, tc, i = _act(env, act)
            if tm["__all__"]:
                break
            kkk = "agent0"
            #assert r[kkk] == 999
            assert i[kkk]["arrive_dest"]
            assert tm[kkk]

            kkk = "agent1"
            #assert r[kkk] != 999
            assert not i[kkk]["arrive_dest"]
            assert not tm[kkk]
            break
    finally:
        env.close()


def test_ma_parking_lot_init_space():
    try:
        for start_seed in [5000, 6000, 7000]:
            for num_agents in [1, 11]:
                for num_others in [0, 2, 4, 8]:
                    for crash_vehicle_penalty in [0, 5]:
                        env_config = dict(
                            start_seed=start_seed,
                            num_agents=num_agents,
                            vehicle_config=dict(lidar=dict(num_others=num_others)),
                            crash_vehicle_penalty=crash_vehicle_penalty
                        )
                        env = MultiAgentParkingLotEnv(env_config)

                        single_space = env.observation_space["agent0"]
                        assert single_space.shape is not None, single_space
                        assert np.prod(single_space.shape) is not None, single_space

                        single_space = env.action_space["agent0"]
                        assert single_space.shape is not None, single_space
                        assert np.prod(single_space.shape) is not None, single_space

                        _check_spaces_before_reset(env)
                        env.reset()
                        _check_spaces_after_reset(env)
                        env.close()
                        # print('Finish: ', env_config)
    finally:
        if "env" in locals():
            env.close()


def test_ma_parking_lot_no_short_episode():
    env = MultiAgentParkingLotEnv({
        "horizon": 300,
        "parking_space_num": 32,
        "num_agents": 35,
    })
    try:
        _check_spaces_before_reset(env)
        o, _ = env.reset()
        _check_spaces_after_reset(env, o)
        actions = [[0, 1], [1, 1], [-1, 1]]
        start = time.time()
        tm_count = 0
        tm = {"__all__": False}
        for step in range(2000):
            # act = {k: actions[np.random.choice(len(actions))] for k in o.keys()}
            act = {k: actions[np.random.choice(len(actions))] for k in env.vehicles.keys()}
            o_keys = set(o.keys()).union({"__all__"})
            a_keys = set(env.action_space.spaces.keys()).union(set(tm.keys()))
            assert o_keys == a_keys
            o, r, tm, tc, i = _act(env, act)
            for kkk, iii in i.items():
                if tm[kkk]:
                    assert iii["episode_length"] >= 1
                    tm_count += 1
            if tm["__all__"]:
                o, _ = env.reset()
                tm = {"__all__": False}
            # if (step + 1) % 100 == 0:
            #     # print(
            #         "Finish {}/2000 simulation steps. Time elapse: {:.4f}. Average FPS: {:.4f}".format(
            #             step + 1,
            #             time.time() - start, (step + 1) / (time.time() - start)
            #         )
            #     )
            if tm_count > 200:
                break
    finally:
        env.close()


def test_ma_parking_lot_horizon_termination():
    # test horizon
    env = MultiAgentParkingLotEnv({"horizon": 100, "num_agents": 8, "crash_done": False})
    try:
        for _ in range(3):  # This function is really easy to break, repeat multiple times!
            _check_spaces_before_reset(env)
            obs, _ = env.reset()
            _check_spaces_after_reset(env, obs)
            assert env.observation_space.contains(obs)
            should_respawn = set()
            for step in range(1, 10000):
                act = {k: [0, 0] for k in env.vehicles.keys()}
                for v_id in act.keys():
                    env.vehicles[v_id].set_static(True)
                obs, r, tm, tc, i = _act(env, act)
                # env.render("top_down", camera_position=(42.5, 0), film_size=(500, 500))
                if step == 0 or step == 1:
                    assert not any(tm.values())
                    assert not any(tc.values())

                if should_respawn:
                    for kkk in should_respawn:
                        assert kkk not in obs, "It seems the max_step agents is not respawn!"
                        assert kkk not in r
                        assert kkk not in tm
                        assert kkk not in tc
                        assert kkk not in i
                    should_respawn.clear()

                for kkk, ddd in tm.items():
                    if ddd and kkk == "__all__":
                        # print("Current: ", step)
                        continue
                    if ddd:
                        assert i[kkk]["max_step"]
                        assert not i[kkk]["out_of_road"]
                        assert not i[kkk]["crash"]
                        assert not i[kkk]["crash_vehicle"]
                        should_respawn.add(kkk)

                if tm["__all__"]:
                    obs, _ = env.reset()
                    should_respawn.clear()
                    break
    finally:
        env.close()


def test_ma_parking_lot_40_agent_reset_after_respawn():
    def check_pos(vehicles):
        while vehicles:
            v_1 = vehicles[0]
            for v_2 in vehicles[1:]:
                v_1_pos = v_1.position
                v_2_pos = v_2.position
                assert norm(
                    v_1_pos[0] - v_2_pos[0], v_1_pos[1] - v_2_pos[1]
                ) > v_1.WIDTH / 2 + v_2.WIDTH / 2, "Vehicles overlap after reset()"
            assert not v_1.crash_vehicle, "Vehicles overlap after reset()"
            vehicles.remove(v_1)

    env = MultiAgentParkingLotEnv({"horizon": 50, "num_agents": 32, "parking_space_num": 32, "use_render": False})
    try:
        _check_spaces_before_reset(env)
        obs, _ = env.reset()
        _check_spaces_after_reset(env, obs)
        assert env.observation_space.contains(obs)
        for step in range(50):
            env.reset()
            check_pos(list(env.vehicles.values()))
            for v_id in list(env.vehicles.keys())[:20]:
                env.agent_manager.finish(v_id)
            env.step({k: [1, 1] for k in env.vehicles.keys()})
            env.step({k: [1, 1] for k in env.vehicles.keys()})
            env.step({k: [1, 1] for k in env.vehicles.keys()})
    finally:
        env.close()


def test_ma_no_reset_error():
    # It is possible that many agents are populated in the same spawn place!
    def check_pos(vehicles):
        while vehicles:
            v_1 = vehicles[0]
            for v_2 in vehicles[1:]:
                v_1_pos = v_1.position
                v_2_pos = v_2.position
                assert norm(
                    v_1_pos[0] - v_2_pos[0], v_1_pos[1] - v_2_pos[1]
                ) > v_1.WIDTH / 2 + v_2.WIDTH / 2, "Vehicles overlap after reset()"
            if v_1.crash_vehicle:
                x = 1
                raise ValueError("Vehicles overlap after reset()")
            vehicles.remove(v_1)

    env = MultiAgentParkingLotEnv({"horizon": 300, "num_agents": 11, "delay_done": 0, "use_render": False})
    try:
        _check_spaces_before_reset(env)
        obs, _ = env.reset()
        _check_spaces_after_reset(env, obs)
        assert env.observation_space.contains(obs)
        for step in range(50):
            check_pos(list(env.vehicles.values()))
            o, r, tm, tc, i = env.step({k: [0, 1] for k in env.vehicles.keys()})
            env.reset()
            if tm["__all__"]:
                break
    finally:
        env.close()


def test_randomize_spawn_place():
    last_pos = {}
    env = MultiAgentParkingLotEnv({"num_agents": 4, "use_render": False, "force_seed_spawn_manager": False})
    try:
        obs, _ = env.reset()
        for step in range(100):
            act = {k: [1, 1] for k in env.vehicles.keys()}
            last_pos = {kkk: v.position for kkk, v in env.vehicles.items()}
            o, r, tm, tc, i = env.step(act)
            obs, _ = env.reset()
            new_pos = {kkk: v.position for kkk, v in env.vehicles.items()}
            for kkk, new_p in new_pos.items():
                assert not np.all(new_p == last_pos[kkk]), (new_p, last_pos[kkk], kkk)
    finally:
        env.close()


if __name__ == '__main__':
    # test_ma_parking_lot_env()
    # test_ma_parking_lot_horizon()
    # test_ma_parking_lot_reset()
    test_ma_parking_lot_reward_done_alignment()
    # test_ma_parking_lot_close_spawn()
    # test_ma_parking_lot_reward_sign()
    # test_ma_parking_lot_init_space()
    # test_ma_parking_lot_no_short_episode()
    # test_ma_parking_lot_horizon_termination()
    # test_ma_parking_lot_40_agent_reset_after_respawn()
    # test_ma_no_reset_error()
    # test_randomize_spawn_place()



# Copyright (c) 2020 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from metadrive.third_party.kitsunetsuki.base.matrices import get_object_matrix

from panda3d.egg import EggVertex, EggVertexUV


class VertexMixin(object):
    def _get_uv_name(self, uv_layer):
        if uv_layer.active:
            return ''
        else:
            return uv_layer.name.replace(' ', '_')

    def make_vertex(self, parent_obj_matrix, obj_matrix, polygon, vertex, use_smooth=False, can_merge=False):
        egg_vertex = EggVertex()
        egg_vertex.set_color((1, 1, 1, 1))

        if can_merge:
            co = obj_matrix @ vertex.co
        else:
            co = vertex.co

        egg_vertex.set_pos(tuple(co))

        if use_smooth:
            normal = parent_obj_matrix @ vertex.normal
        else:
            normal = parent_obj_matrix @ polygon.normal

        egg_vertex.set_normal(tuple(normal))

        return egg_vertex

    def make_vertex_uv(self, uv_layer, uv):
        egg_vertex_uv = EggVertexUV(self._get_uv_name(uv_layer), tuple(uv))
        return egg_vertex_uv



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from __future__ import division

import collections

from metadrive.render_pipeline.rpcore.globals import Globals
from metadrive.render_pipeline.rpcore.pluginbase.base_plugin import BasePlugin

from .voxelization_stage import VoxelizationStage
from .vxgi_stage import VXGIStage


class Plugin(BasePlugin):

    name = "Voxel Global Illumination"
    author = "tobspr <tobias.springer1@gmail.com>"
    description = (
        "Provides Global Illumination using Voxel Cone Tracing. This "
        "technique is still very unoptimized and experimental!"
    )
    version = "alpha (!)"

    def on_stage_setup(self):
        self._voxel_stage = self.create_stage(VoxelizationStage)
        self._vxgi_stage = self.create_stage(VXGIStage)

        self._voxel_stage.voxel_resolution = self.get_setting("grid_resolution")
        self._voxel_stage.voxel_world_size = self.get_setting("grid_ws_size")

        if self.is_plugin_enabled("pssm"):
            # Add shadow map as requirement
            self._voxel_stage.required_pipes.append("PSSMSceneSunShadowMapPCF")
            self._voxel_stage.required_inputs.append("PSSMSceneSunShadowMVP")

    def on_pre_render_update(self):
        task = self._queue[0]
        self._queue.rotate(-1)
        task()

    def on_pipeline_created(self):
        self._queue = collections.deque()
        self._queue.extend([self._voxelize_x, self._voxelize_y, self._voxelize_z])
        self._queue.extend([self._generate_mipmaps])

    def _set_grid_pos(self):
        """ Finds the new voxel grid position """
        grid_pos = Globals.base.camera.get_pos(Globals.base.render)

        # Snap the voxel grid
        voxel_size = 2.0 * self.get_setting("grid_ws_size") / self.get_setting("grid_resolution")
        snap_size = voxel_size * 2**4

        for dimension in range(3):
            cell_val = grid_pos.get_cell(dimension)
            grid_pos.set_cell(dimension, cell_val - cell_val % snap_size)

        self._voxel_stage.set_grid_position(grid_pos)

    def _voxelize_x(self):
        """ Voxelizes the scene from the x axis """
        self._set_grid_pos()
        self._voxel_stage.state = VoxelizationStage.S_voxelize_x

    def _voxelize_y(self):
        """ Voxelizes the scene from the y axis """
        self._voxel_stage.state = VoxelizationStage.S_voxelize_y

    def _voxelize_z(self):
        """ Voxelizes the scene from the z axis """
        self._voxel_stage.state = VoxelizationStage.S_voxelize_z

    def _generate_mipmaps(self):
        """ Generates the mipmaps for the voxel grid """
        self._voxel_stage.state = VoxelizationStage.S_gen_mipmaps



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from panda3d.core import Vec2, TransparencyAttrib, Vec3
from direct.gui.DirectFrame import DirectFrame
from direct.gui.DirectButton import DirectButton
from direct.gui.DirectGui import DGG

from metadrive.render_pipeline.rpcore.rpobject import RPObject
from metadrive.render_pipeline.rpcore.globals import Globals
from metadrive.render_pipeline.rpcore.gui.text import Text


class DraggableWindow(RPObject):
    """ This is a simple draggable but not resizeable window """
    def __init__(self, width=800, height=500, title="Window", parent=None):
        """ Constructs a new window with the given dimensions and title """
        RPObject.__init__(self, "Window-" + title)
        self._width = width
        self._height = height
        self._title = title
        self._visible = True
        self._parent = parent if parent else Globals.base.pixel2d
        self._dragging = False
        self._drag_offset = Vec2(0)
        self._pos = Vec2(0)

    def center_on_screen(self):
        """ Centers the window on screen """
        self._context_scale = 1.0 / self._parent.get_sx()
        self._context_width = Globals.native_resolution.x * self._context_scale
        self._context_height = Globals.native_resolution.y * self._context_scale
        self._set_pos(Vec2((self._context_width - self._width) / 2, (self._context_height - self._height) / 2))

    def set_title(self, title):
        """ Sets the window title """
        self._title = title
        self._window_title.set_text(title)

    def show(self):
        """ Shows the window """
        self._visible = True
        self.center_on_screen()
        self._node.show()

    def hide(self):
        """ Hides the window """
        self._visible = False
        self._stop_drag()
        self._node.hide()

    def remove(self):
        """ Removes the window from the scene graph. You should still delete the
        instance """
        self._stop_drag()
        self._node.remove_node()

    def _create_components(self):
        """ Creates the window components """
        self._node = self._parent.attach_new_node("Window")
        self._node.set_pos(self._pos.x, 1, -self._pos.y)
        border_px = 1
        border_frame_size = (-border_px, self._width + border_px, border_px, -self._height - border_px)
        self._border_frame = DirectFrame(
            pos=(0, 1, 0),
            frameSize=border_frame_size,
            frameColor=(24 / 255.0, 131 / 255.0, 215 / 255.0, 1),
            parent=self._node,
            state=DGG.NORMAL
        )
        self._background = DirectFrame(
            pos=(0, 1, 0),
            frameSize=(0, self._width, 0, -self._height),
            frameColor=(0.1, 0.1, 0.1, 1.0),
            parent=self._node
        )
        self._title_bar = DirectFrame(
            pos=(0, 1, 0),
            frameSize=(0, self._width, 0, -25),
            # frameColor=(0.058, 0.058, 0.058, 1),
            frameColor=(1, 1, 1, 1),
            parent=self._node,
            state=DGG.NORMAL
        )
        self._window_title = Text(
            parent=self._node, x=8, y=17, text=self._title, size=13, color=Vec3(0.15), may_change=True
        )
        self._btn_close = DirectButton(
            relief=DGG.FLAT,
            pressEffect=1,
            pos=(self._width - 22, 1, -12),
            frameColor=(1.0, 0.2, 0.2, 0.5),
            parent=self._node,
            scale=(45 / 2, 1, 24 / 2),
            image="/$$rp/data/gui/close_window.png"
        )

        # Init bindings
        self._btn_close.set_transparency(TransparencyAttrib.M_alpha)
        self._btn_close.bind(DGG.B1CLICK, self._request_close)
        self._btn_close.bind(DGG.WITHIN, self._on_close_btn_hover)
        self._btn_close.bind(DGG.WITHOUT, self._on_close_btn_out)
        self._title_bar.bind(DGG.B1PRESS, self._start_drag)
        self._title_bar.bind(DGG.B1RELEASE, self._stop_drag)

    def _start_drag(self, evt=None):  # pylint: disable=unused-argument
        """ Gets called when the user starts dragging the window """
        self._dragging = True
        self._node.detach_node()
        self._node.reparent_to(self._parent)
        Globals.base.taskMgr.add(self._on_tick, "UIWindowDrag", uponDeath=self._stop_drag)
        self._drag_offset = self._pos - self._get_mouse_pos()

    def _on_close_btn_hover(self, evt=None):  # pylint: disable=unused-argument
        """ Internal method when the close button got hovered """
        self._btn_close["image"] = "/$$rp/data/gui/close_window_hover.png"

    def _on_close_btn_out(self, evt=None):  # pylint: disable=unused-argument
        """ Internal method when the close button is no longer hovered """
        self._btn_close["image"] = "/$$rp/data/gui/close_window.png"

    def _request_close(self, evt=None):  # pylint: disable=unused-argument
        """ This method gets called when the close button gets clicked """
        self.hide()

    def _stop_drag(self, evt=None):  # pylint: disable=unused-argument
        """ Gets called when the user stops dragging the window """
        Globals.base.taskMgr.remove("UIWindowDrag")
        self._dragging = False

    def _get_mouse_pos(self):
        """ Internal helper function to get the mouse position, scaled by
        the context scale """
        mouse_x, mouse_y = (Globals.base.win.get_pointer(0).x, Globals.base.win.get_pointer(0).y)
        return Vec2(mouse_x, mouse_y) * self._context_scale

    def _set_pos(self, pos):
        """ Moves the window to the specified position """
        self._pos = pos
        self._pos.x = max(self._pos.x, -self._width + 100)
        self._pos.y = max(self._pos.y, 25)
        self._pos.x = min(self._pos.x, self._context_width - 100)
        self._pos.y = min(self._pos.y, self._context_height - 50)
        self._node.set_pos(self._pos.x, 1, -self._pos.y)

    def _on_tick(self, task):
        """ Task which updates the window while being dragged """
        self._set_pos(self._get_mouse_pos() + self._drag_offset)
        return task.cont



import copy
from collections import deque

import numpy as np

from metadrive.component.lane.straight_lane import StraightLane
from metadrive.component.pgblock.create_pg_block_utils import CreateAdverseRoad, CreateRoadFrom, ExtendStraightLane, \
    create_bend_straight
from metadrive.component.pgblock.pg_block import PGBlock, PGBlockSocket
from metadrive.component.road_network import Road
from metadrive.constants import PGLineType
from metadrive.utils.pg.utils import check_lane_on_road
from metadrive.component.pg_space import ParameterSpace, Parameter, BlockParameterSpace


class InterSection(PGBlock):
    """
                                up(Goal:1)
                                   ||
                                   ||
                                   ||
                                   ||
                                   ||
                  _________________||_________________
    left(Goal:2)  -----------------||---------------- right(Goal:0)
                               __  ||
                              |    ||
             spawn_road    <--|    ||
                              |    ||
                              |__  ||
                                  down
    It's an Intersection with two lanes on same direction, 4 lanes on both roads
    """

    ID = "X"
    EXTRA_PART = "extra"
    PARAMETER_SPACE = ParameterSpace(BlockParameterSpace.INTERSECTION)
    SOCKET_NUM = 3
    ANGLE = 90  # may support other angle in the future
    EXIT_PART_LENGTH = 35

    _enable_u_turn_flag = False

    # LEFT_TURN_NUM = 1 now it is useless

    def __init__(self, *args, **kwargs):
        if "radius" in kwargs:
            self.radius = kwargs.pop("radius")
        else:
            self.radius = None
        super(InterSection, self).__init__(*args, **kwargs)
        if self.radius is None:
            self.radius = self.get_config(copy=False)[Parameter.radius]

    def _try_plug_into_previous_block(self) -> bool:
        para = self.get_config()
        decrease_increase = -1 if para[Parameter.decrease_increase] == 0 else 1
        if self.positive_lane_num <= 1:
            decrease_increase = 1
        elif self.positive_lane_num >= 4:
            decrease_increase = -1
        self.lane_num_intersect = self.positive_lane_num + decrease_increase * para[Parameter.change_lane_num]
        no_cross = True
        attach_road = self.pre_block_socket.positive_road
        _attach_road = self.pre_block_socket.negative_road
        attach_lanes = attach_road.get_lanes(self._global_network)
        # right straight left node name, rotate it to fit different part
        intersect_nodes = deque(
            [self.road_node(0, 0),
             self.road_node(1, 0),
             self.road_node(2, 0), _attach_road.start_node]
        )

        for i in range(4):
            right_lane, success = self._create_part(attach_lanes, attach_road, self.radius, intersect_nodes, i)
            no_cross = no_cross and success
            if i != 3:
                lane_num = self.positive_lane_num if i == 1 else self.lane_num_intersect
                exit_road = Road(self.road_node(i, 0), self.road_node(i, 1))
                no_cross = CreateRoadFrom(
                    right_lane,
                    lane_num,
                    exit_road,
                    self.block_network,
                    self._global_network,
                    ignore_intersection_checking=self.ignore_intersection_checking
                ) and no_cross
                no_cross = CreateAdverseRoad(
                    exit_road,
                    self.block_network,
                    self._global_network,
                    ignore_intersection_checking=self.ignore_intersection_checking
                ) and no_cross
                socket = PGBlockSocket(exit_road, -exit_road)
                self.add_respawn_roads(socket.negative_road)
                self.add_sockets(socket)
                attach_road = -exit_road
                attach_lanes = attach_road.get_lanes(self.block_network)
        return no_cross

    def _create_part(self, attach_lanes, attach_road: Road, radius: float, intersect_nodes: deque,
                     part_idx) -> (StraightLane, bool):
        lane_num = self.lane_num_intersect if part_idx == 0 or part_idx == 2 else self.positive_lane_num
        non_cross = True
        attach_left_lane = attach_lanes[0]
        # first left part
        assert isinstance(attach_left_lane, StraightLane), "Can't create a intersection following a circular lane"
        self._create_left_turn(radius, lane_num, attach_left_lane, attach_road, intersect_nodes, part_idx)

        # u-turn
        if self._enable_u_turn_flag:
            adverse_road = -attach_road
            self._create_u_turn(attach_road, part_idx)

        # go forward part
        lanes_on_road = copy.copy(attach_lanes)
        straight_lane_len = 2 * radius + (2 * lane_num - 1) * lanes_on_road[0].width_at(0)
        for l in lanes_on_road:
            next_lane = ExtendStraightLane(l, straight_lane_len, (PGLineType.NONE, PGLineType.NONE))
            self.block_network.add_lane(attach_road.end_node, intersect_nodes[1], next_lane)

        # right part
        length = self.EXIT_PART_LENGTH
        right_turn_lane = lanes_on_road[-1]
        assert isinstance(right_turn_lane, StraightLane), "Can't create a intersection following a circular lane"
        right_bend, right_straight = create_bend_straight(
            right_turn_lane, length, radius, np.deg2rad(self.ANGLE), True, right_turn_lane.width_at(0),
            (PGLineType.NONE, PGLineType.SIDE)
        )

        non_cross = (
            not check_lane_on_road(
                self._global_network, right_bend, 1, ignore_intersection_checking=self.ignore_intersection_checking
            )
        ) and non_cross
        CreateRoadFrom(
            right_bend,
            min(self.positive_lane_num, self.lane_num_intersect),
            Road(attach_road.end_node, intersect_nodes[0]),
            self.block_network,
            self._global_network,
            toward_smaller_lane_index=True,
            side_lane_line_type=PGLineType.SIDE,
            inner_lane_line_type=PGLineType.NONE,
            center_line_type=PGLineType.NONE,
            ignore_intersection_checking=self.ignore_intersection_checking
        )

        intersect_nodes.rotate(-1)
        right_straight.line_types = [PGLineType.BROKEN, PGLineType.SIDE]
        return right_straight, non_cross

    def get_socket(self, index: int) -> PGBlockSocket:
        socket = super(InterSection, self).get_socket(index)
        if socket.negative_road in self.get_respawn_roads():
            self._respawn_roads.remove(socket.negative_road)
        return socket

    def _create_left_turn(self, radius, lane_num, attach_left_lane, attach_road, intersect_nodes, part_idx):
        left_turn_radius = radius + lane_num * attach_left_lane.width_at(0)
        diff = self.lane_num_intersect - self.positive_lane_num  # increase lane num
        if ((part_idx == 1 or part_idx == 3) and diff > 0) or ((part_idx == 0 or part_idx == 2) and diff < 0):
            diff = abs(diff)
            left_bend, extra_part = create_bend_straight(
                attach_left_lane, self.lane_width * diff, left_turn_radius, np.deg2rad(self.ANGLE), False,
                attach_left_lane.width_at(0), (PGLineType.NONE, PGLineType.NONE)
            )
            left_road_start = intersect_nodes[2]
            pre_left_road_start = left_road_start + self.EXTRA_PART
            CreateRoadFrom(
                left_bend,
                min(self.positive_lane_num, self.lane_num_intersect),
                Road(attach_road.end_node, pre_left_road_start),
                self.block_network,
                self._global_network,
                toward_smaller_lane_index=False,
                center_line_type=PGLineType.NONE,
                side_lane_line_type=PGLineType.NONE,
                inner_lane_line_type=PGLineType.NONE,
                ignore_intersection_checking=self.ignore_intersection_checking
            )

            CreateRoadFrom(
                extra_part,
                min(self.positive_lane_num, self.lane_num_intersect),
                Road(pre_left_road_start, left_road_start),
                self.block_network,
                self._global_network,
                toward_smaller_lane_index=False,
                center_line_type=PGLineType.NONE,
                side_lane_line_type=PGLineType.NONE,
                inner_lane_line_type=PGLineType.NONE,
                ignore_intersection_checking=self.ignore_intersection_checking
            )

        else:
            left_bend, _ = create_bend_straight(
                attach_left_lane, self.EXIT_PART_LENGTH, left_turn_radius, np.deg2rad(self.ANGLE), False,
                attach_left_lane.width_at(0), (PGLineType.NONE, PGLineType.NONE)
            )
            left_road_start = intersect_nodes[2]
            CreateRoadFrom(
                left_bend,
                min(self.positive_lane_num, self.lane_num_intersect),
                Road(attach_road.end_node, left_road_start),
                self.block_network,
                self._global_network,
                toward_smaller_lane_index=False,
                center_line_type=PGLineType.NONE,
                side_lane_line_type=PGLineType.NONE,
                inner_lane_line_type=PGLineType.NONE,
                ignore_intersection_checking=self.ignore_intersection_checking
            )

    def _create_u_turn(self, attach_road, part_idx):
        # set to CONTINUOUS to debug
        line_type = PGLineType.NONE
        lanes = attach_road.get_lanes(self.block_network) if part_idx != 0 else self.positive_lanes
        attach_left_lane = lanes[0]
        lane_num = len(lanes)
        left_turn_radius = self.lane_width / 2
        left_bend, _ = create_bend_straight(
            attach_left_lane, 0.1, left_turn_radius, np.deg2rad(180), False, attach_left_lane.width_at(0),
            (PGLineType.NONE, PGLineType.NONE)
        )
        left_road_start = (-attach_road).start_node
        CreateRoadFrom(
            left_bend,
            lane_num,
            Road(attach_road.end_node, left_road_start),
            self.block_network,
            self._global_network,
            toward_smaller_lane_index=False,
            center_line_type=line_type,
            side_lane_line_type=line_type,
            inner_lane_line_type=line_type,
            ignore_intersection_checking=self.ignore_intersection_checking
        )

    def enable_u_turn(self, enable_u_turn: bool):
        self._enable_u_turn_flag = enable_u_turn

    def get_intermediate_spawn_lanes(self):
        """Override this function for intersection so that we won't spawn vehicles in the center of intersection."""
        respawn_lanes = self.get_respawn_lanes()
        return respawn_lanes



import copy
import logging
from collections import namedtuple
from typing import Dict

import math
import numpy as np

from metadrive.component.lane.abs_lane import AbstractLane
from metadrive.component.map.base_map import BaseMap
from metadrive.component.road_network import Road
from metadrive.component.vehicle.base_vehicle import BaseVehicle
from metadrive.constants import TARGET_VEHICLES, TRAFFIC_VEHICLES, OBJECT_TO_AGENT, AGENT_TO_OBJECT
from metadrive.manager.base_manager import BaseManager
from metadrive.utils import merge_dicts

BlockVehicles = namedtuple("block_vehicles", "trigger_road vehicles")


class TrafficMode:
    # Traffic vehicles will be respawned, once they arrive at the destinations
    Respawn = "respawn"

    # Traffic vehicles will be triggered only once
    Trigger = "trigger"

    # Hybrid, some vehicles are triggered once on map and disappear when arriving at destination, others exist all time
    Hybrid = "hybrid"


class PGTrafficManager(BaseManager):
    VEHICLE_GAP = 10  # m

    def __init__(self):
        """
        Control the whole traffic flow
        """
        super(PGTrafficManager, self).__init__()

        self._traffic_vehicles = []

        # triggered by the event. TODO(lqy) In the future, event trigger can be introduced
        self.block_triggered_vehicles = []

        # traffic property
        self.mode = self.engine.global_config["traffic_mode"]
        self.random_traffic = self.engine.global_config["random_traffic"]
        self.density = self.engine.global_config["traffic_density"]
        self.respawn_lanes = None

    def reset(self):
        """
        Generate traffic on map, according to the mode and density
        :return: List of Traffic vehicles
        """
        map = self.current_map
        logging.debug("load scene {}".format("Use random traffic" if self.random_traffic else ""))

        # update vehicle list
        self.block_triggered_vehicles = []

        traffic_density = self.density
        if abs(traffic_density) < 1e-2:
            return
        self.respawn_lanes = self._get_available_respawn_lanes(map)
        if self.mode == TrafficMode.Respawn:
            # add respawn vehicle
            self._create_respawn_vehicles(map, traffic_density)
        elif self.mode == TrafficMode.Trigger or self.mode == TrafficMode.Hybrid:
            self._create_vehicles_once(map, traffic_density)
        else:
            raise ValueError("No such mode named {}".format(self.mode))

    def before_step(self):
        """
        All traffic vehicles make driving decision here
        :return: None
        """
        # trigger vehicles
        engine = self.engine
        if self.mode != TrafficMode.Respawn:
            for v in engine.agent_manager.active_agents.values():
                ego_lane_idx = v.lane_index[:-1]
                ego_road = Road(ego_lane_idx[0], ego_lane_idx[1])
                if len(self.block_triggered_vehicles) > 0 and \
                        ego_road == self.block_triggered_vehicles[-1].trigger_road:
                    block_vehicles = self.block_triggered_vehicles.pop()
                    self._traffic_vehicles += list(self.get_objects(block_vehicles.vehicles).values())
        for v in self._traffic_vehicles:
            p = self.engine.get_policy(v.name)
            v.before_step(p.act())
        return dict()

    def after_step(self, *args, **kwargs):
        """
        Update all traffic vehicles' states,
        """
        v_to_remove = []
        for v in self._traffic_vehicles:
            v.after_step()
            if not v.on_lane:
                if self.mode == TrafficMode.Trigger:
                    v_to_remove.append(v)
                elif self.mode == TrafficMode.Respawn or self.mode == TrafficMode.Hybrid:
                    v_to_remove.append(v)
                else:
                    raise ValueError("Traffic mode error: {}".format(self.mode))
        for v in v_to_remove:
            vehicle_type = type(v)
            self.clear_objects([v.id])
            self._traffic_vehicles.remove(v)
            if self.mode == TrafficMode.Respawn or self.mode == TrafficMode.Hybrid:
                lane = self.respawn_lanes[self.np_random.randint(0, len(self.respawn_lanes))]
                lane_idx = lane.index
                long = self.np_random.rand() * lane.length / 2
                traffic_v_config = {"spawn_lane_index": lane_idx, "spawn_longitude": long}
                new_v = self.spawn_object(vehicle_type, vehicle_config=traffic_v_config)
                from metadrive.policy.idm_policy import IDMPolicy
                self.add_policy(new_v.id, IDMPolicy, new_v, self.generate_seed())
                self._traffic_vehicles.append(new_v)

        return dict()

    def before_reset(self) -> None:
        """
        Clear the scene and then reset the scene to empty
        :return: None
        """
        super(PGTrafficManager, self).before_reset()
        self.density = self.engine.global_config["traffic_density"]
        self.block_triggered_vehicles = []
        self._traffic_vehicles = []

    def get_vehicle_num(self):
        """
        Get the vehicles on road
        :return:
        """
        if self.mode == TrafficMode.Respawn:
            return len(self._traffic_vehicles)
        return sum(len(block_vehicle_set.vehicles) for block_vehicle_set in self.block_triggered_vehicles)

    def get_global_states(self) -> Dict:
        """
        Return all traffic vehicles' states
        :return: States of all vehicles
        """
        states = dict()
        traffic_states = dict()
        for vehicle in self._traffic_vehicles:
            traffic_states[vehicle.index] = vehicle.get_state()

        # collect other vehicles
        if self.mode != TrafficMode.Respawn:
            for v_b in self.block_triggered_vehicles:
                for vehicle in v_b.vehicles:
                    traffic_states[vehicle.index] = vehicle.get_state()
        states[TRAFFIC_VEHICLES] = traffic_states
        active_obj = copy.copy(self.engine.agent_manager._active_objects)
        pending_obj = copy.copy(self.engine.agent_manager._pending_objects)
        dying_obj = copy.copy(self.engine.agent_manager._dying_objects)
        states[TARGET_VEHICLES] = {k: v.get_state() for k, v in active_obj.items()}
        states[TARGET_VEHICLES] = merge_dicts(
            states[TARGET_VEHICLES], {k: v.get_state()
                                      for k, v in pending_obj.items()}, allow_new_keys=True
        )
        states[TARGET_VEHICLES] = merge_dicts(
            states[TARGET_VEHICLES], {k: v_count[0].get_state()
                                      for k, v_count in dying_obj.items()},
            allow_new_keys=True
        )

        states[OBJECT_TO_AGENT] = copy.deepcopy(self.engine.agent_manager._object_to_agent)
        states[AGENT_TO_OBJECT] = copy.deepcopy(self.engine.agent_manager._agent_to_object)
        return states

    def get_global_init_states(self) -> Dict:
        """
        Special handling for first states of traffic vehicles
        :return: States of all vehicles
        """
        vehicles = dict()
        for vehicle in self._traffic_vehicles:
            init_state = vehicle.get_state()
            init_state["index"] = vehicle.index
            init_state["type"] = vehicle.class_name
            init_state["enable_respawn"] = vehicle.enable_respawn
            vehicles[vehicle.index] = init_state

        # collect other vehicles
        if self.mode != TrafficMode.Respawn:
            for v_b in self.block_triggered_vehicles:
                for vehicle in v_b.vehicles:
                    init_state = vehicle.get_state()
                    init_state["type"] = vehicle.class_name
                    init_state["index"] = vehicle.index
                    init_state["enable_respawn"] = vehicle.enable_respawn
                    vehicles[vehicle.index] = init_state
        return vehicles

    def _propose_vehicle_configs(self, lane: AbstractLane):
        potential_vehicle_configs = []
        total_num = int(lane.length / self.VEHICLE_GAP)
        vehicle_longs = [i * self.VEHICLE_GAP for i in range(total_num)]
        # Only choose given number of vehicles
        for long in vehicle_longs:
            random_vehicle_config = {"spawn_lane_index": lane.index, "spawn_longitude": long, "enable_reverse": False}
            potential_vehicle_configs.append(random_vehicle_config)
        return potential_vehicle_configs

    def _create_respawn_vehicles(self, map: BaseMap, traffic_density: float):
        total_num = len(self.respawn_lanes)
        for lane in self.respawn_lanes:
            _traffic_vehicles = []
            total_num = int(lane.length / self.VEHICLE_GAP)
            vehicle_longs = [i * self.VEHICLE_GAP for i in range(total_num)]
            self.np_random.shuffle(vehicle_longs)
            for long in vehicle_longs[:int(np.ceil(traffic_density * len(vehicle_longs)))]:
                # if self.np_random.rand() > traffic_density and abs(lane.length - InRampOnStraight.RAMP_LEN) > 0.1:
                #     # Do special handling for ramp, and there must be vehicles created there
                #     continue
                vehicle_type = self.random_vehicle_type()
                traffic_v_config = {"spawn_lane_index": lane.index, "spawn_longitude": long}
                traffic_v_config.update(self.engine.global_config["traffic_vehicle_config"])
                random_v = self.spawn_object(vehicle_type, vehicle_config=traffic_v_config)
                from metadrive.policy.idm_policy import IDMPolicy
                self.add_policy(random_v.id, IDMPolicy, random_v, self.generate_seed())
                self._traffic_vehicles.append(random_v)

    def _create_vehicles_once(self, map: BaseMap, traffic_density: float) -> None:
        """
        Trigger mode, vehicles will be triggered only once, and disappear when arriving destination
        :param map: Map
        :param traffic_density: it can be adjusted each episode
        :return: None
        """
        vehicle_num = 0
        for block in map.blocks[1:]:

            # Propose candidate locations for spawning new vehicles
            trigger_lanes = block.get_intermediate_spawn_lanes()
            if self.engine.global_config["need_inverse_traffic"] and block.ID in ["S", "C", "r", "R"]:
                neg_lanes = block.block_network.get_negative_lanes()
                self.np_random.shuffle(neg_lanes)
                trigger_lanes += neg_lanes
            potential_vehicle_configs = []
            for lanes in trigger_lanes:
                for l in lanes:
                    if hasattr(self.engine, "object_manager") and l in self.engine.object_manager.accident_lanes:
                        continue
                    potential_vehicle_configs += self._propose_vehicle_configs(l)

            # How many vehicles should we spawn in this block?
            total_length = sum([lane.length for lanes in trigger_lanes for lane in lanes])
            total_spawn_points = int(math.floor(total_length / self.VEHICLE_GAP))
            total_vehicles = int(math.floor(total_spawn_points * traffic_density))

            # Generate vehicles!
            vehicles_on_block = []
            self.np_random.shuffle(potential_vehicle_configs)
            selected = potential_vehicle_configs[:min(total_vehicles, len(potential_vehicle_configs))]
            # print("We have {} candidates! We are spawning {} vehicles!".format(total_vehicles, len(selected)))

            from metadrive.policy.idm_policy import IDMPolicy
            for v_config in selected:
                vehicle_type = self.random_vehicle_type()
                v_config.update(self.engine.global_config["traffic_vehicle_config"])
                random_v = self.spawn_object(vehicle_type, vehicle_config=v_config)
                self.add_policy(random_v.id, IDMPolicy, random_v, self.generate_seed())
                vehicles_on_block.append(random_v.name)

            trigger_road = block.pre_block_socket.positive_road
            block_vehicles = BlockVehicles(trigger_road=trigger_road, vehicles=vehicles_on_block)

            self.block_triggered_vehicles.append(block_vehicles)
            vehicle_num += len(vehicles_on_block)
        self.block_triggered_vehicles.reverse()

    def _get_available_respawn_lanes(self, map: BaseMap) -> list:
        """
        Used to find some respawn lanes
        :param map: select spawn lanes from this map
        :return: respawn_lanes
        """
        respawn_lanes = []
        respawn_roads = []
        for block in map.blocks:
            roads = block.get_respawn_roads()
            for road in roads:
                if road in respawn_roads:
                    respawn_roads.remove(road)
                else:
                    respawn_roads.append(road)
        for road in respawn_roads:
            respawn_lanes += road.get_lanes(map.road_network)
        return respawn_lanes

    def random_vehicle_type(self):
        from metadrive.component.vehicle.vehicle_type import random_vehicle_type
        vehicle_type = random_vehicle_type(self.np_random, [0.2, 0.3, 0.3, 0.2, 0.0, 0.0, 0.0])
        return vehicle_type

    def destroy(self) -> None:
        """
        Destory func, release resource
        :return: None
        """
        self.clear_objects([v.id for v in self._traffic_vehicles])
        self._traffic_vehicles = []
        # current map

        # traffic vehicle list
        self._traffic_vehicles = None
        self.block_triggered_vehicles = None

        # traffic property
        self.mode = None
        self.random_traffic = None
        self.density = None

    def __del__(self):
        logging.debug("{} is destroyed".format(self.__class__.__name__))

    def __repr__(self):
        return self.vehicles.__repr__()

    @property
    def vehicles(self):
        return list(self.engine.get_objects(filter=lambda o: isinstance(o, BaseVehicle)).values())

    @property
    def traffic_vehicles(self):
        return list(self._traffic_vehicles)

    def seed(self, random_seed):
        if not self.random_traffic:
            super(PGTrafficManager, self).seed(random_seed)

    @property
    def current_map(self):
        return self.engine.map_manager.current_map

    def get_state(self):
        ret = super(PGTrafficManager, self).get_state()
        ret["_traffic_vehicles"] = [v.name for v in self._traffic_vehicles]
        flat = []
        for b_v in self.block_triggered_vehicles:
            flat.append((b_v.trigger_road.start_node, b_v.trigger_road.end_node, b_v.vehicles))
        ret["block_triggered_vehicles"] = flat
        return ret

    def set_state(self, state: dict, old_name_to_current=None):
        super(PGTrafficManager, self).set_state(state, old_name_to_current)
        self._traffic_vehicles = list(
            self.get_objects([old_name_to_current[name] for name in state["_traffic_vehicles"]]).values()
        )
        self.block_triggered_vehicles = [
            BlockVehicles(trigger_road=Road(s, e), vehicles=[old_name_to_current[name] for name in v])
            for s, e, v in state["block_triggered_vehicles"]
        ]


# For compatibility check
TrafficManager = PGTrafficManager


class MixedPGTrafficManager(PGTrafficManager):
    def _create_respawn_vehicles(self, *args, **kwargs):
        raise NotImplementedError()

    def _create_vehicles_once(self, map: BaseMap, traffic_density: float) -> None:
        vehicle_num = 0
        for block in map.blocks[1:]:

            # Propose candidate locations for spawning new vehicles
            trigger_lanes = block.get_intermediate_spawn_lanes()
            if self.engine.global_config["need_inverse_traffic"] and block.ID in ["S", "C", "r", "R"]:
                neg_lanes = block.block_network.get_negative_lanes()
                self.np_random.shuffle(neg_lanes)
                trigger_lanes += neg_lanes
            potential_vehicle_configs = []
            for lanes in trigger_lanes:
                for l in lanes:
                    if hasattr(self.engine, "object_manager") and l in self.engine.object_manager.accident_lanes:
                        continue
                    potential_vehicle_configs += self._propose_vehicle_configs(l)

            # How many vehicles should we spawn in this block?
            total_length = sum([lane.length for lanes in trigger_lanes for lane in lanes])
            total_spawn_points = int(math.floor(total_length / self.VEHICLE_GAP))
            total_vehicles = int(math.floor(total_spawn_points * traffic_density))

            # Generate vehicles!
            vehicles_on_block = []
            self.np_random.shuffle(potential_vehicle_configs)
            selected = potential_vehicle_configs[:min(total_vehicles, len(potential_vehicle_configs))]

            from metadrive.policy.idm_policy import IDMPolicy
            from metadrive.policy.expert_policy import ExpertPolicy
            # print("===== We are initializing {} vehicles =====".format(len(selected)))
            # print("Current seed: ", self.engine.global_random_seed)
            for v_config in selected:
                vehicle_type = self.random_vehicle_type()
                v_config.update(self.engine.global_config["traffic_vehicle_config"])
                random_v = self.spawn_object(vehicle_type, vehicle_config=v_config)
                if self.np_random.random() < self.engine.global_config["rl_agent_ratio"]:
                    # print("Vehicle {} is assigned with RL policy!".format(random_v.id))
                    self.add_policy(random_v.id, ExpertPolicy, random_v, self.generate_seed())
                else:
                    self.add_policy(random_v.id, IDMPolicy, random_v, self.generate_seed())
                vehicles_on_block.append(random_v.name)

            trigger_road = block.pre_block_socket.positive_road
            block_vehicles = BlockVehicles(trigger_road=trigger_road, vehicles=vehicles_on_block)

            self.block_triggered_vehicles.append(block_vehicles)
            vehicle_num += len(vehicles_on_block)
        self.block_triggered_vehicles.reverse()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'point_insert.ui'
#
# Created by: PyQt5 UI code generator 5.7
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(342, 197)
        Dialog.setMinimumSize(QtCore.QSize(342, 197))
        Dialog.setMaximumSize(QtCore.QSize(342, 197))
        self.label = QtWidgets.QLabel(Dialog)
        self.label.setGeometry(QtCore.QRect(10, 10, 301, 71))
        self.label.setWordWrap(True)
        self.label.setObjectName("label")
        self.label_2 = QtWidgets.QLabel(Dialog)
        self.label_2.setGeometry(QtCore.QRect(10, 100, 46, 21))
        self.label_2.setObjectName("label_2")
        self.label_3 = QtWidgets.QLabel(Dialog)
        self.label_3.setGeometry(QtCore.QRect(10, 130, 46, 21))
        self.label_3.setObjectName("label_3")
        self.ipt_value = QtWidgets.QDoubleSpinBox(Dialog)
        self.ipt_value.setGeometry(QtCore.QRect(70, 130, 121, 22))
        self.ipt_value.setMaximum(99999.99)
        self.ipt_value.setObjectName("ipt_value")
        self.ipt_time = QtWidgets.QTimeEdit(Dialog)
        self.ipt_time.setGeometry(QtCore.QRect(70, 100, 121, 22))
        self.ipt_time.setObjectName("ipt_time")
        self.btn_insert = QtWidgets.QPushButton(Dialog)
        self.btn_insert.setGeometry(QtCore.QRect(240, 160, 75, 23))
        self.btn_insert.setFocusPolicy(QtCore.Qt.TabFocus)
        self.btn_insert.setObjectName("btn_insert")

        self.retranslateUi(Dialog)
        QtCore.QMetaObject.connectSlotsByName(Dialog)
        Dialog.setTabOrder(self.ipt_time, self.ipt_value)
        Dialog.setTabOrder(self.ipt_value, self.btn_insert)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Insert Point"))
        self.label.setText(
            _translate(
                "Dialog", "Enter the values of the point you want to insert. \n"
                "\n"
                "Did you know: If you dont want to enter concrete values, just click and drag anywhere on the curve to insert a new point."
            )
        )
        self.label_2.setText(_translate("Dialog", "Time:"))
        self.label_3.setText(_translate("Dialog", "Value:"))
        self.ipt_time.setDisplayFormat(_translate("Dialog", "HH:mm"))
        self.btn_insert.setText(_translate("Dialog", "Insert"))


from . import resources_rc



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from __future__ import division
from functools import partial

from panda3d.core import Texture, Vec3

from direct.gui.DirectFrame import DirectFrame
from direct.gui.DirectGui import DGG
from direct.gui.DirectScrolledFrame import DirectScrolledFrame

from metadrive.render_pipeline.rplibs.six import itervalues

from metadrive.render_pipeline.rpcore.image import Image
from metadrive.render_pipeline.rpcore.util.display_shader_builder import DisplayShaderBuilder
from metadrive.render_pipeline.rpcore.globals import Globals
from metadrive.render_pipeline.rpcore.render_target import RenderTarget
from metadrive.render_pipeline.rpcore.gui.texture_preview import TexturePreview
from metadrive.render_pipeline.rpcore.gui.sprite import Sprite
from metadrive.render_pipeline.rpcore.gui.labeled_checkbox import LabeledCheckbox
from metadrive.render_pipeline.rpcore.gui.text import Text
from metadrive.render_pipeline.rpcore.gui.draggable_window import DraggableWindow


class BufferViewer(DraggableWindow):
    """ This class provides a view into the buffers to inspect them """
    def __init__(self, pipeline, parent):
        """ Constructs the buffer viewer """
        DraggableWindow.__init__(self, width=1400, height=800, parent=parent, title="Buffer- and Image-Browser")
        self._pipeline = pipeline
        self._scroll_height = 3000
        self._display_images = False
        self._stages = []
        self._create_components()
        self._tex_preview = TexturePreview(self._pipeline, parent)
        self._tex_preview.hide()
        self.hide()

    def toggle(self):
        """ Updates all the buffers and then toggles the buffer viewer """
        if self._visible:
            self._remove_components()
            self.hide()
        else:
            self._perform_update()
            self.show()

    @property
    def entries(self):
        """ Returns a list of all registered entries """
        return RenderTarget.REGISTERED_TARGETS + Image.REGISTERED_IMAGES

    @property
    def stage_information(self):
        """ Returns the amount of attached stages, and also the memory consumed
        in MiB in a tuple. """
        count, memory = 0, 0
        for entry in self.entries:
            if isinstance(entry, Texture):
                memory += entry.estimate_texture_memory()
                count += 1
            elif entry.__class__.__name__ == "RenderTarget":
                for target in itervalues(entry.targets):
                    memory += target.estimate_texture_memory()
                    count += 1
            else:
                self.warn("Unkown type:", entry.__class__.__name__)
        return memory, count

    def _create_components(self):
        """ Creates the window components """
        DraggableWindow._create_components(self)

        self._content_frame = DirectScrolledFrame(
            frameSize=(0, self._width - 15, 0, self._height - 70),
            canvasSize=(0, self._width - 80, 0, self._scroll_height),
            autoHideScrollBars=False,
            scrollBarWidth=12.0,
            frameColor=(0, 0, 0, 0),
            verticalScroll_relief=DGG.FLAT,
            verticalScroll_incButton_relief=DGG.FLAT,
            verticalScroll_decButton_relief=DGG.FLAT,
            verticalScroll_thumb_relief=DGG.FLAT,
            verticalScroll_frameColor=(0.05, 0.05, 0.05, 1),
            verticalScroll_thumb_frameColor=(0.8, 0.8, 0.8, 1),
            verticalScroll_incButton_frameColor=(0.6, 0.6, 0.6, 1),
            verticalScroll_decButton_frameColor=(0.6, 0.6, 0.6, 1),
            horizontalScroll_frameColor=(0, 0, 0, 0),
            horizontalScroll_relief=False,
            horizontalScroll_thumb_relief=False,
            horizontalScroll_incButton_relief=False,
            horizontalScroll_decButton_relief=False,
            parent=self._node,
            pos=(0, 1, -self._height)
        )
        self._content_node = self._content_frame.getCanvas().attach_new_node("BufferComponents")
        self._content_node.set_scale(1, 1, -1)
        self._content_node.set_z(self._scroll_height)
        self._chb_show_images = LabeledCheckbox(
            parent=self._node,
            x=10,
            y=43,
            chb_callback=self._set_show_images,
            chb_checked=False,
            text="Display image resources",
            text_color=Vec3(0.4),
            expand_width=330
        )

    def _set_show_images(self, arg):
        """ Sets whether images and textures will be shown """
        self._display_images = arg
        self._perform_update()

    def _set_scroll_height(self, height):
        """ Sets the maximum scroll height in the content frame """
        self._scroll_height = height
        self._content_frame["canvasSize"] = (0, self._width - 80, 0, self._scroll_height)
        self._content_node.set_z(self._scroll_height)

    def _remove_components(self):
        """ Removes all components of the buffer viewer """
        self._content_node.node().remove_all_children()
        self._tex_preview.hide()

    def _perform_update(self):
        """ Collects all entries, extracts their images and re-renders the
        window """

        # Collect texture stages
        self._stages = []
        for entry in sorted(self.entries, key=lambda entry: entry.sort):
            if isinstance(entry, Texture):
                if self._display_images:
                    self._stages.append(entry)
            # Can not use isinstance or we get circular import references
            elif entry.__class__.__name__ == "RenderTarget":
                for target in itervalues(entry.targets):
                    self._stages.append(target)
            else:
                self.warn("Unrecognized instance!", entry.__class__)

        self._render_stages()

    def _on_texture_hovered(self, hover_frame, evt=None):  # pylint: disable=unused-argument
        """ Internal method when a texture is hovered """
        hover_frame["frameColor"] = (0, 0, 0, 0.1)

    def _on_texture_blurred(self, hover_frame, evt=None):  # pylint: disable=unused-argument
        """ Internal method when a texture is blurred """
        hover_frame["frameColor"] = (0, 0, 0, 0)

    def _on_texture_clicked(self, tex_handle, evt=None):  # pylint: disable=unused-argument
        """ Internal method when a texture is blurred """
        self._tex_preview.present(tex_handle)

    def _render_stages(self):
        """ Renders the stages to the window """

        self._remove_components()
        entries_per_row = 6
        aspect = Globals.native_resolution.y / Globals.native_resolution.x
        entry_width = 235
        entry_height = (entry_width - 20) * aspect + 55

        # Store already processed images
        processed = set()
        index = -1
        # Iterate over all stages
        for stage_tex in self._stages:
            if stage_tex in processed:
                continue
            processed.add(stage_tex)
            index += 1
            stage_name = stage_tex.get_name()

            xoffs = index % entries_per_row
            yoffs = index // entries_per_row
            node = self._content_node.attach_new_node("Preview")
            node.set_sz(-1)
            node.set_pos(10 + xoffs * (entry_width - 14), 1, yoffs * (entry_height - 14 + 10))

            r, g, b = 0.2, 0.2, 0.2
            if isinstance(stage_tex, Image):
                r, g, b = 0.2, 0.4, 0.6

            stage_name = stage_name.replace("render_pipeline_internal:", "")
            parts = stage_name.split(":")
            stage_name = parts[-1]
            DirectFrame(
                parent=node,
                frameSize=(7, entry_width - 17, -7, -entry_height + 17),
                frameColor=(r, g, b, 1.0),
                pos=(0, 0, 0)
            )

            frame_hover = DirectFrame(
                parent=node,
                frameSize=(0, entry_width - 10, 0, -entry_height + 10),
                frameColor=(0, 0, 0, 0),
                pos=(0, 0, 0),
                state=DGG.NORMAL
            )
            frame_hover.bind(DGG.ENTER, partial(self._on_texture_hovered, frame_hover))
            frame_hover.bind(DGG.EXIT, partial(self._on_texture_blurred, frame_hover))
            frame_hover.bind(DGG.B1PRESS, partial(self._on_texture_clicked, stage_tex))

            Text(text=stage_name, x=15, y=29, parent=node, size=12, color=Vec3(0.8))

            # Scale image so it always fits
            w, h = stage_tex.get_x_size(), stage_tex.get_y_size()
            padd_x, padd_y = 24, 57
            scale_x = (entry_width - padd_x) / max(1, w)
            scale_y = (entry_height - padd_y) / max(1, h)
            scale_factor = min(scale_x, scale_y)

            if stage_tex.get_texture_type() == Image.TT_buffer_texture:
                scale_factor = 1
                w = entry_width - padd_x
                h = entry_height - padd_y

            preview = Sprite(
                image=stage_tex,
                w=scale_factor * w,
                h=scale_factor * h,
                any_filter=False,
                parent=node,
                x=7,
                y=40,
                transparent=False
            )

            preview.set_shader_inputs(mipmap=0, slice=0, brightness=1, tonemap=False)

            preview_shader = DisplayShaderBuilder.build(stage_tex, scale_factor * w, scale_factor * h)
            preview.set_shader(preview_shader)

        num_rows = (index + entries_per_row) // entries_per_row

        self._set_scroll_height(50 + (entry_height - 14 + 10) * num_rows)



from panda3d.core import SamplerState, Shader, ConfigVariableString, Texture

from metadrive.base_class.base_object import BaseObject
from metadrive.constants import CamMask
from metadrive.engine.asset_loader import AssetLoader
from metadrive.utils.utils import is_mac


class SkyBox(BaseObject):
    """
    SkyBox is only related to render
    """
    ROTATION_MAX = 5000

    def __init__(self, pure_background: bool = False):
        super(SkyBox, self).__init__(random_seed=0)
        self._accumulate = 0
        self.f = 1
        if not self.render or pure_background:
            return
        skybox = self.loader.loadModel(AssetLoader.file_path("models", "skybox.bam"))

        skybox.hide(CamMask.MiniMap | CamMask.Shadow | CamMask.ScreenshotCam)
        skybox.set_scale(20000)

        skybox_texture = self.loader.loadTexture(AssetLoader.file_path("textures", "skybox.jpg"))
        skybox_texture.set_format(Texture.F_srgb)
        skybox_texture.set_minfilter(SamplerState.FT_linear)
        skybox_texture.set_magfilter(SamplerState.FT_linear)
        skybox_texture.set_wrap_u(SamplerState.WM_repeat)
        skybox_texture.set_wrap_v(SamplerState.WM_mirror)
        skybox_texture.set_anisotropic_degree(16)
        skybox.set_texture(skybox_texture)

        gles = ConfigVariableString("load-display").getValue()
        if gles == "pandagles2":
            skybox_shader = Shader.load(
                Shader.SL_GLSL, AssetLoader.file_path("shaders", "skybox_gles.vert.glsl"),
                AssetLoader.file_path("shaders", "skybox_gles.frag.glsl")
            )
        else:
            if is_mac():
                vert_file = "skybox_mac.vert.glsl"
                frag_file = "skybox_mac.frag.glsl"
            else:
                vert_file = "skybox.vert.glsl"
                frag_file = "skybox.frag.glsl"
            skybox_shader = Shader.load(
                Shader.SL_GLSL, AssetLoader.file_path("shaders", vert_file),
                AssetLoader.file_path("shaders", frag_file)
            )
        skybox.set_shader(skybox_shader)
        skybox.reparentTo(self.origin)
        skybox.setZ(-4400)
        skybox.setH(0)

    def step(self):
        if not self.render:
            return
        if self._accumulate >= self.ROTATION_MAX:
            self.f *= -1
            self._accumulate = 0
        self._accumulate += 1
        factor = self.f * (1 - abs(self._accumulate - self.ROTATION_MAX / 2) * 2 / self.ROTATION_MAX)
        self.set_heading_theta(self.origin.getH() + factor * 0.0035, in_rad=False)



import copy
import logging

from metadrive.policy.idm_policy import TrajectoryIDMPOlicy
from typing import Dict

from gymnasium.spaces import Box, Dict, MultiDiscrete, Discrete

from metadrive.constants import DEFAULT_AGENT
from metadrive.manager.base_manager import BaseManager
from metadrive.policy.AI_protect_policy import AIProtectPolicy
from metadrive.policy.manual_control_policy import ManualControlPolicy
from metadrive.policy.replay_policy import ReplayTrafficParticipantPolicy


class AgentManager(BaseManager):
    """
    This class maintain the relationship between active agents in the environment with the underlying instance
    of objects.

    Note:
    agent name: Agent name that exists in the environment, like agent0, agent1, ....
    object name: The unique name for each object, typically be random string.
    """
    INITIALIZED = False  # when vehicles instances are created, it will be set to True

    def __init__(self, init_observations, init_action_space):
        """
        The real init is happened in self.init(), in which super().__init__() will be called
        """
        # BaseVehicles which can be controlled by policies when env.step() called
        self._active_objects = {}
        # BaseVehicles which will be recycled after the delay_done time
        self._dying_objects = {}
        self._agents_finished_this_frame = dict()  # for observation space

        self.next_agent_count = 0

        # fake init. before creating engine and vehicles, it is necessary when all vehicles re-created in runtime
        self.observations = copy.copy(init_observations)  # its value is map<agent_id, obs> before init() is called
        self._init_observations = init_observations  # map <agent_id, observation>
        # init spaces before initializing env.engine
        observation_space = {
            agent_id: single_obs.observation_space
            for agent_id, single_obs in init_observations.items()
        }
        assert isinstance(init_action_space, dict)
        assert isinstance(observation_space, dict)
        self._init_observation_spaces = observation_space
        self._init_action_spaces = init_action_space
        self.observation_spaces = copy.copy(observation_space)
        self.action_spaces = copy.copy(init_action_space)
        self.episode_created_agents = None

        # this map will be override when the env.init() is first called and vehicles are made
        self._agent_to_object = {k: k for k in self.observations.keys()}  # no target vehicles created, fake init
        self._object_to_agent = {k: k for k in self.observations.keys()}  # no target vehicles created, fake init

        # get the value in init()
        self._allow_respawn = None
        self._debug = None
        self._delay_done = None
        self._infinite_agents = None

    def _get_vehicles(self, config_dict: dict):
        from metadrive.component.vehicle.vehicle_type import random_vehicle_type, vehicle_type
        ret = {}
        for agent_id, v_config in config_dict.items():
            v_type = random_vehicle_type(self.np_random) if self.engine.global_config["random_agent_model"] else \
                vehicle_type[v_config["vehicle_model"] if v_config.get("vehicle_model", False) else "default"]

            obj_name = agent_id if self.engine.global_config["force_reuse_object_name"] else None
            obj = self.spawn_object(v_type, vehicle_config=v_config, name=obj_name)
            ret[agent_id] = obj
            policy_cls = self.agent_policy
            args = [obj, self.generate_seed()]
            if policy_cls == TrajectoryIDMPOlicy or issubclass(policy_cls, TrajectoryIDMPOlicy):
                args.append(self.engine.map_manager.current_sdc_route)
            self.add_policy(obj.id, policy_cls, *args)
        return ret

    @property
    def agent_policy(self):
        # note: agent.id = object id

        if self.engine.global_config["manual_control"]:
            if self.engine.global_config.get("use_AI_protector", False):
                policy = AIProtectPolicy
            else:
                policy = ManualControlPolicy
        else:
            policy = self.engine.global_config["agent_policy"]
        return policy

    def before_reset(self):
        if not self.INITIALIZED:
            super(AgentManager, self).__init__()
            self.INITIALIZED = True

        self.episode_created_agents = None

        if not self.engine.replay_episode:
            for v in self.dying_agents.values():
                self._remove_vehicle(v)

        for v in self.get_vehicle_list():
            if hasattr(v, "before_reset"):
                v.before_reset()

        super(AgentManager, self).before_reset()

    def reset(self):
        """
        Agent manager is really initialized after the BaseVehicle Instances are created
        """
        self.random_spawn_lane_in_single_agent()
        config = self.engine.global_config
        self._debug = config["debug"]
        self._delay_done = config["delay_done"]
        self._infinite_agents = config["num_agents"] == -1
        self._allow_respawn = config["allow_respawn"]
        self.episode_created_agents = self._get_vehicles(
            config_dict=self.engine.global_config["target_vehicle_configs"]
        )

    def after_reset(self):
        init_vehicles = self.episode_created_agents
        vehicles_created = set(init_vehicles.keys())
        vehicles_in_config = set(self._init_observations.keys())
        assert vehicles_created == vehicles_in_config, "{} not defined in target vehicles config".format(
            vehicles_created.difference(vehicles_in_config)
        )

        # it is used when reset() is called to reset its original agent_id
        self._agent_to_object = {agent_id: vehicle.name for agent_id, vehicle in init_vehicles.items()}
        self._object_to_agent = {vehicle.name: agent_id for agent_id, vehicle in init_vehicles.items()}
        self._active_objects = {v.name: v for v in init_vehicles.values()}
        self._dying_objects = {}
        self._agents_finished_this_frame = dict()

        # real init {obj_name: space} map
        self.observations = dict()
        self.observation_spaces = dict()
        self.action_spaces = dict()
        for agent_id, vehicle in init_vehicles.items():
            self.observations[vehicle.name] = self._init_observations[agent_id]
            obs_space = self._init_observation_spaces[agent_id]
            self.observation_spaces[vehicle.name] = obs_space
            if not self.engine.global_config["image_observation"]:
                assert isinstance(obs_space, Box)
            else:
                assert isinstance(obs_space, Dict), "Multi-agent observation should be gym.Dict"
            action_space = self._init_action_spaces[agent_id]
            self.action_spaces[vehicle.name] = action_space
            assert isinstance(action_space, Box) or isinstance(action_space,
                                                               MultiDiscrete) or isinstance(action_space, Discrete)
        self.next_agent_count = len(init_vehicles)

    def set_state(self, state: dict, old_name_to_current=None):
        super(AgentManager, self).set_state(state, old_name_to_current)
        created_agents = state["created_agents"]
        created_agents = {agent_id: old_name_to_current[obj_name] for agent_id, obj_name in created_agents.items()}
        episode_created_agents = {}
        for a_id, name in created_agents.items():
            episode_created_agents[a_id] = self.engine.get_objects([name])[name]
        self.episode_created_agents = episode_created_agents

    def get_state(self):
        ret = super(AgentManager, self).get_state()
        agent_info = {agent_id: obj.name for agent_id, obj in self.episode_created_agents.items()}
        ret["created_agents"] = agent_info
        return ret

    def random_spawn_lane_in_single_agent(self):
        if not self.engine.global_config["is_multi_agent"] and \
                self.engine.global_config.get("random_spawn_lane_index", False) and self.engine.current_map is not None:
            spawn_road_start = self.engine.global_config["target_vehicle_configs"][DEFAULT_AGENT]["spawn_lane_index"][0]
            spawn_road_end = self.engine.global_config["target_vehicle_configs"][DEFAULT_AGENT]["spawn_lane_index"][1]
            index = self.np_random.randint(self.engine.current_map.config["lane_num"])
            self.engine.global_config["target_vehicle_configs"][DEFAULT_AGENT]["spawn_lane_index"] = (
                spawn_road_start, spawn_road_end, index
            )

    def finish(self, agent_name, ignore_delay_done=False):
        """
        ignore_delay_done: Whether to ignore the delay done. This is not required when the agent success the episode!
        """
        if not self.engine.replay_episode:
            vehicle_name = self._agent_to_object[agent_name]
            v = self._active_objects.pop(vehicle_name)
            if (not ignore_delay_done) and (self._delay_done > 0):
                self._put_to_dying_queue(v)
            else:
                # move to invisible place
                self._remove_vehicle(v)
            self._agents_finished_this_frame[agent_name] = v.name
            self._check()

    def _check(self):
        if self._debug:
            current_keys = sorted(list(self._active_objects.keys()) + list(self._dying_objects.keys()))
            exist_keys = sorted(list(self._object_to_agent.keys()))
            assert current_keys == exist_keys, "You should confirm_respawn() after request for propose_new_vehicle()!"

    def propose_new_vehicle(self):
        # Create a new vehicle.
        agent_name = self.next_agent_id()
        next_config = self.engine.global_config["target_vehicle_configs"]["agent0"]
        vehicle = self._get_vehicles({agent_name: next_config})[agent_name]
        new_v_name = vehicle.name
        self._agent_to_object[agent_name] = new_v_name
        self._object_to_agent[new_v_name] = agent_name
        # TODO: this may cause error? Sharing observation
        logging.warning("Test MARL new agent observation to avoid bug!")
        self.observations[new_v_name] = self._init_observations["agent0"]
        self.observations[new_v_name].reset(vehicle)
        self.observation_spaces[new_v_name] = self._init_observation_spaces["agent0"]
        self.action_spaces[new_v_name] = self._init_action_spaces["agent0"]
        self._active_objects[vehicle.name] = vehicle
        self._check()
        step_info = vehicle.before_step([0, 0])
        vehicle.set_static(False)
        return agent_name, vehicle, step_info

    def next_agent_id(self):
        ret = "agent{}".format(self.next_agent_count)
        self.next_agent_count += 1
        return ret

    def set_allow_respawn(self, flag: bool):
        self._allow_respawn = flag

    def try_actuate_agent(self, step_infos, stage="before_step"):
        """
        Some policies should make decision before physics world actuation, in particular, those need decision-making
        But other policies like ReplayPolicy should be called in after_step, as they already know the final state and
        exempt the requirement for rolling out the dynamic system to get it.
        """
        assert stage == "before_step" or stage == "after_step"
        for agent_id in self.active_agents.keys():
            policy = self.get_policy(self._agent_to_object[agent_id])
            cond_1 = stage == "before_step" and not isinstance(policy, ReplayTrafficParticipantPolicy)
            cond_2 = stage == "after_step" and isinstance(policy, ReplayTrafficParticipantPolicy)
            if cond_2 or cond_1:
                assert policy is not None, "No policy is set for agent {}".format(agent_id)
                action = policy.act(agent_id)
                step_infos[agent_id] = policy.get_action_info()
                step_infos[agent_id].update(self.get_agent(agent_id).before_step(action))
        return step_infos

    def before_step(self):
        # not in replay mode
        self._agents_finished_this_frame = dict()
        step_infos = self.try_actuate_agent(dict(), stage="before_step")

        finished = set()
        for v_name in self._dying_objects.keys():
            self._dying_objects[v_name][1] -= 1
            if self._dying_objects[v_name][1] <= 0:  # Countdown goes to 0, it's time to remove the vehicles!
                v = self._dying_objects[v_name][0]
                self._remove_vehicle(v)
                finished.add(v_name)
        for v_name in finished:
            self._dying_objects.pop(v_name)
        return step_infos

    def after_step(self, *args, **kwargs):
        step_infos = self.for_each_active_agents(lambda v: v.after_step())
        step_infos = self.try_actuate_agent(step_infos, stage="after_step")
        return step_infos

    def _translate(self, d):
        return {self._object_to_agent[k]: v for k, v in d.items()}

    def get_vehicle_list(self):
        return list(self._active_objects.values()) + [v for (v, _) in self._dying_objects.values()]

    def get_observations(self):
        if hasattr(self, "engine") and self.engine.replay_episode:
            return self.engine.replay_manager.get_replay_agent_observations()
        else:
            ret = {
                old_agent_id: self.observations[v_name]
                for old_agent_id, v_name in self._agents_finished_this_frame.items()
            }
            for obj_id, observation in self.observations.items():
                if self.is_active_object(obj_id):
                    ret[self.object_to_agent(obj_id)] = observation
            return ret

    def get_observation_spaces(self):
        ret = {
            old_agent_id: self.observation_spaces[v_name]
            for old_agent_id, v_name in self._agents_finished_this_frame.items()
        }
        for obj_id, space in self.observation_spaces.items():
            if self.is_active_object(obj_id):
                ret[self.object_to_agent(obj_id)] = space
        return ret

    def get_action_spaces(self):
        ret = dict()
        for obj_id, space in self.action_spaces.items():
            if self.is_active_object(obj_id):
                ret[self.object_to_agent(obj_id)] = space
        return ret

    def is_active_object(self, object_name):
        if not self.INITIALIZED:
            return True
        return True if object_name in self._active_objects.keys() else False

    @property
    def active_agents(self):
        """
        Return Map<agent_id, BaseVehicle>
        """
        if hasattr(self, "engine") and self.engine is not None and self.engine.replay_episode:
            return self.engine.replay_manager.replay_agents
        else:
            return {self._object_to_agent[k]: v for k, v in self._active_objects.items()}

    @property
    def dying_agents(self):
        assert not self.engine.replay_episode
        return {self._object_to_agent[k]: v for k, (v, _) in self._dying_objects.items()}

    @property
    def just_terminated_agents(self):
        assert not self.engine.replay_episode
        ret = {}
        for agent_name, v_name in self._agents_finished_this_frame.items():
            v = self.get_object(v_name, raise_error=False)
            ret[agent_name] = v
        return ret

    @property
    def active_objects(self):
        """
        Return meta-data, a pointer, Caution !
        :return: Map<obj_name, obj>
        """
        raise DeprecationWarning("prohibit! Use active agent instead")
        return self._active_objects

    def get_agent(self, agent_name):
        object_name = self.agent_to_object(agent_name)
        return self.get_object(object_name)

    def get_object(self, object_name, raise_error=True):
        if object_name in self._active_objects:
            return self._active_objects[object_name]
        elif object_name in self._dying_objects:
            return self._dying_objects[object_name][0]
        else:
            if raise_error:
                raise ValueError("Object {} not found!".format(object_name))
            else:
                return None

    def object_to_agent(self, obj_name):
        """
        We recommend to use engine.agent_to_object() or engine.object_to_agent() instead of the ones in agent_manager,
        since this two functions DO NOT work when replaying episode.
        :param obj_name: BaseVehicle name
        :return: agent id
        """
        # if obj_name not in self._active_objects.keys() and self.INITIALIZED:
        #     raise ValueError("You can not access a pending Object(BaseVehicle) outside the agent_manager!")
        return self._object_to_agent[obj_name]

    def agent_to_object(self, agent_id):
        """
        We recommend to use engine.agent_to_object() or engine.object_to_agent() instead of the ones in agent_manager,
        since this two functions DO NOT work when replaying episode.
        """
        return self._agent_to_object[agent_id]

    def destroy(self):
        # when new agent joins in the game, we only change this two maps.
        if self.INITIALIZED:
            super(AgentManager, self).destroy()
        self._agent_to_object = {}
        self._object_to_agent = {}

        # BaseVehicles which can be controlled by policies when env.step() called
        self._active_objects = {}

        # BaseVehicles which can be respawned
        self._dying_objects = {}

        # Dict[object_id: value], init for **only** once after spawning vehicle
        self.observations = {}
        self.observation_spaces = {}
        self.action_spaces = {}

        self.next_agent_count = 0
        self.INITIALIZED = False

    def _put_to_dying_queue(self, v, ignore_delay_done=False):
        vehicle_name = v.name
        v.set_static(True)
        self._dying_objects[vehicle_name] = [v, 0 if ignore_delay_done else self._delay_done]

    def _remove_vehicle(self, vehicle):
        vehicle_name = vehicle.name
        assert vehicle_name not in self._active_objects
        self.clear_objects([vehicle_name])
        self._agent_to_object.pop(self._object_to_agent[vehicle_name])
        self._object_to_agent.pop(vehicle_name)

    @property
    def allow_respawn(self):
        if not self._allow_respawn:
            return False
        if len(self._active_objects) + len(self._dying_objects) < self.engine.global_config["num_agents"] \
                or self._infinite_agents:
            return True
        else:
            return False

    def for_each_active_agents(self, func, *args, **kwargs):
        """
        This func is a function that take each vehicle as the first argument and *arg and **kwargs as others.
        """
        assert len(self.active_agents) > 0, "Not enough vehicles exist!"
        ret = dict()
        for k, v in self.active_agents.items():
            ret[k] = func(v, *args, **kwargs)
        return ret



# Copyright (c) 2020 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import bpy
import configparser
import os

from bpy_extras.io_utils import ExportHelper
from typing import Set, cast

from ..gltf import GLTFExporter
from ..gltf import spec
from .armature import ArmatureMixin

BLENDSHAPE_PRESETS = (
    'neutral',
    'a',
    'i',
    'u',
    'e',
    'o',
    'blink',
    'joy',
    'angry',
    'sorrow',
    'fun',
    'lookup',
    'lookdown',
    'lookleft',
    'lookright',
    'blink_l',
    'blink_r',
)


class VRMExporter(ArmatureMixin, GLTFExporter):
    def __init__(self, args):
        super().__init__(args)

        self._z_up = False
        self._pose_freeze = True
        self._export_type = 'all'

    def _add_vrm_thumbnail(self, gltf_node, filepath):
        gltf_sampler = {
            'name': os.path.basename(filepath),
            'wrapS': spec.CLAMP_TO_EDGE,
            'wrapT': spec.CLAMP_TO_EDGE,
        }
        gltf_node['samplers'].append(gltf_sampler)

        gltf_image = {
            'name': os.path.basename(filepath),
            'mimeType': 'image/png',
            'extras': {
                'uri': filepath,
            }
        }
        gltf_node['images'].append(gltf_image)

        gltf_texture = {
            'sampler': len(gltf_node['samplers']) - 1,
            'source': len(gltf_node['images']) - 1,
        }
        gltf_node['textures'].append(gltf_texture)

        texid = len(gltf_node['textures']) - 1
        gltf_node['extensions']['VRM']['meta']['texture'] = texid

    def make_root_node(self):
        gltf_node = super().make_root_node()

        data = {}
        text = bpy.data.texts.get('VRM.ini')
        if text:
            data = configparser.ConfigParser()
            data.read_string(text.as_string())
        else:
            raise RuntimeError('Missing "VRM.ini" text block.')

        vrm_meta = {
            'exporterVersion': gltf_node['asset']['generator'],
            'specVersion': '0.0',
            'meta': {
                'title': data['meta']['title'],
                'version': data['meta']['version'],
                'author': data['meta']['author'],
                'contactInformation': data['meta']['contactInformation'],
                'reference': data['meta']['reference'],
                'texture': 0,  # thumbnail texture
                'allowedUserName': data['meta'].get('allowedUserName', 'OnlyAuthor'),
                'violentUssageName': data['meta'].get('violentUssageName', 'Disallow'),
                'sexualUssageName': data['meta'].get('sexualUssageName', 'Disallow'),
                'commercialUssageName': data['meta'].get('commercialUssageName', 'Disallow'),
                'otherPermissionUrl': data['meta'].get('otherPermissionUrl', ''),
                'licenseName': data['meta'].get('licenseName', 'Redistribution_Prohibited'),
                'otherLicenseUrl': data['meta'].get('otherLicenseUrl', ''),
            },
            'humanoid': {
                'armStretch': 0.0,
                'legStretch': 0.0,
                'lowerArmTwist': 0.0,  # LowerArm bone roll
                'upperArmTwist': 0.0,  # UpperArm bone roll
                'lowerLegTwist': 0.0,  # LowerLeg bone roll
                'upperLegTwist': 0.0,  # UpperLeg bone roll
                'feetSpacing': 0.0,
                'hasTranslationDoF': False,
                'humanBones': [],
            },
            'firstPerson': {
                'firstPersonBone': None,
                'firstPersonBoneOffset': {
                    'x': 0,
                    'y': 0,
                    'z': 0,
                },
                'meshAnnotations': [],
                'lookAtTypeName': 'Bone',
                # 'lookAtTypeName': 'BlendShape',
                'lookAtHorizontalInner': None,
                'lookAtHorizontalOuter': None,
                'lookAtVerticalDown': None,
                'lookAtVerticalUp': None,
            },
            'blendShapeMaster': {
                'blendShapeGroups': [],
            },
            'secondaryAnimation': {
                'boneGroups': [],
                # 'colliderGroups': [],
            },
            'materialProperties': [],
        }

        gltf_node['extensionsUsed'].append('VRM')
        gltf_node['extensions']['VRM'] = vrm_meta
        gltf_node['materials'] = []

        # make thumbnail
        if self._inputs:
            prefix = os.path.basename(self._inputs[0]).replace('.blend', '.png')
            inpdir = os.path.dirname(os.path.abspath(self._inputs[0]))
            if os.path.exists(inpdir) and os.path.isdir(inpdir):
                for filename in reversed(sorted(os.listdir(inpdir))):
                    if filename.startswith(prefix):
                        self._add_vrm_thumbnail(gltf_node, os.path.join(inpdir, filename))
                        break

        return gltf_node

    def _make_vrm_material(self, material):
        vrm_material = {
            'floatProperties': {
                '_BlendMode': 0 if material.blend_method == 'OPAQUE' else 1,
                '_BumpScale': 1,
                '_CullMode': 2 if material.use_backface_culling else 0,
                '_Cutoff': material.alpha_threshold,
                '_DebugMode': 0,
                '_DstBlend': 0,
                '_IndirectLightIntensity': 0.1,
                '_LightColorAttenuation': 0,
                '_MToonVersion': 35,
                '_OutlineColorMode': 0,
                '_OutlineCullMode': 1,
                '_OutlineLightingMix': 1,
                '_OutlineScaledMaxDistance': 1,
                '_OutlineWidth': 0.5,
                '_OutlineWidthMode': 0,
                '_ReceiveShadowRate': 1,
                '_RimFresnelPower': 1,
                '_RimLift': 0,
                '_RimLightingMix': 0,
                '_ShadeShift': 0,
                '_ShadeToony': 0.9,
                '_ShadingGradeRate': 1,
                '_SrcBlend': 1,
                '_UvAnimRotation': 0,
                '_UvAnimScrollX': 0,
                '_UvAnimScrollY': 0,
                '_ZWrite': 1,
            },
            'keywordMap': {},
            'name': material.name,
            'renderQueue': 2000,
            'shader': 'VRM_USE_GLTFSHADER',
            'tagMap': {},
            'textureProperties': {},
            'vectorProperties': {
                '_BumpMap': [0, 0, 1, 1],
                '_Color': [1, 1, 1, 1],
                '_EmissionColor': [0, 0, 0, 1],
                '_EmissionMap': [0, 0, 1, 1],
                '_MainTex': [0, 0, 1, 1],
                '_OutlineColor': [0, 0, 0, 1],
                '_OutlineWidthTexture': [0, 0, 1, 1],
                '_ReceiveShadowTexture': [0, 0, 1, 1],
                '_RimColor': [0, 0, 0, 1],
                '_RimTexture': [0, 0, 1, 1],
                '_ShadeColor': [1, 1, 1, 1],
                '_ShadeTexture': [0, 0, 1, 1],
                '_ShadingGradeTexture': [0, 0, 1, 1],
                '_SphereAdd': [0, 0, 1, 1],
                '_UvAnimMaskTexture': [0, 0, 1, 1],
            },
        }

        return vrm_material

    def _make_vrm_blend_shape(self, name):
        """
        Standby expression:
        - Neutral

        Lip-sync:
        - A (aa)
        - I (ih)
        - U (ou)
        - E (e)
        - O (oh)

        Blink:
        - Blink
        - Blink_L
        - Blink_R

        Emotion:
        - Fun
        - Angry
        - Sorrow
        - Joy

        Eye control:
        - LookUp
        - LookDown
        - LookLeft
        - LookRight
        """

        vrm_name = {
            # try to get VRM blend shapes from VRChat
            'vrc.v_aa': 'A',
            'vrc.v_ih': 'I',
            'vrc.v_ou': 'U',
            'vrc.v_e': 'E',
            'vrc.v_oh': 'O',
            'vrc.blink': 'Blink',
            'vrc.blink_left': 'Blink_L',
            'vrc.blink_right': 'Blink_R',
        }.get(name, name)

        preset = 'unknown'
        if vrm_name.lower() in BLENDSHAPE_PRESETS:
            preset = vrm_name.lower()

        vrm_blend_shape = {
            'name': vrm_name,
            'presetName': preset,
            'isBinary': False,
            'binds': [],  # bind to mesh ID and shape key ID with shape key weight
            'materialValues': [],  # material values override
        }

        return vrm_blend_shape

    def convert(self):
        root, buffer_ = super().convert()

        for gltf_material_id, gltf_material in enumerate(root['materials']):
            material = bpy.data.materials[gltf_material['name']]
            vrm_material = self._make_vrm_material(material)

            if gltf_material['alphaMode'] == 'OPAQUE':
                vrm_material['tagMap']['RenderType'] = 'Opaque'
                vrm_material['shader'] = 'VRM/MToon'
            else:
                vrm_material['shader'] = 'VRM/UnlitCutout'

            if gltf_material['pbrMetallicRoughness'].get('baseColorTexture'):
                vrm_material['textureProperties']['_MainTex'] = gltf_material['pbrMetallicRoughness']['baseColorTexture'
                                                                                                      ]['index']

            root['extensions']['VRM']['materialProperties'].append(vrm_material)

        vrm_blend_shapes = {
            'Neutral': {
                'name': 'Neutral',
                'presetName': 'neutral',
                'isBinary': False,
                'binds': [],
                'materialValues': [],
            }
        }
        for gltf_mesh_id, gltf_mesh in enumerate(root['meshes']):
            vrm_annotation = {
                'firstPersonFlag': 'Auto',
                'mesh': gltf_mesh_id,
            }
            root['extensions']['VRM']['firstPerson']['meshAnnotations'].append(vrm_annotation)

            for gltf_primitive_id, gltf_primitive in enumerate(gltf_mesh['primitives']):
                for sk_id, sk_name in enumerate(gltf_primitive['extras']['targetNames']):
                    if sk_name in vrm_blend_shapes:
                        vrm_blend_shape = vrm_blend_shapes[sk_name]
                    else:
                        vrm_blend_shape = self._make_vrm_blend_shape(sk_name)
                        vrm_blend_shapes[sk_name] = vrm_blend_shape

                    for vrm_bind in vrm_blend_shape['binds']:
                        if vrm_bind['mesh'] == gltf_mesh_id and vrm_bind['index'] == sk_id:
                            break
                    else:
                        vrm_bind = {
                            'mesh': gltf_mesh_id,
                            'index': sk_id,
                            'weight': 100,
                        }
                        vrm_blend_shape['binds'].append(vrm_bind)

        for vrm_blend_shape in vrm_blend_shapes.values():
            root['extensions']['VRM']['blendShapeMaster']['blendShapeGroups'].append(vrm_blend_shape)

        return root, buffer_


class VRMExporterOperator(bpy.types.Operator, ExportHelper):
    bl_idname = 'avatar.vrm'
    bl_label = 'Export VRM'
    bl_description = 'Export VRM'
    bl_options = {'REGISTER', 'UNDO'}

    filename_ext = '.vrm'
    filter_glob: bpy.props.StringProperty(default='*.vrm', options={'HIDDEN'})

    def execute(self, context: bpy.types.Context):
        if not self.filepath:
            return {'CANCELLED'}

        class Args(object):
            inputs = []
            output = self.filepath
            export = 'all'
            render = 'default'
            exec = None
            action = None
            speed = None
            scale = None
            merge = None
            keep = None
            no_extra_uv = None
            no_materials = None
            no_textures = None
            empty_textures = None
            set_origin = None
            normalize_weights = None

        args = Args()
        e = VRMExporter(args)
        out, buf = e.convert()

        e.write(out, args.output, is_binary=True)

        # re-open current file
        bpy.ops.wm.open_mainfile(filepath=bpy.data.filepath)

        return {"FINISHED"}

    def invoke(self, context, event):
        return cast(Set[str], ExportHelper.invoke(self, context, event))

    def draw(self, context):
        pass


def export(export_op, context):
    export_op.layout.operator(VRMExporterOperator.bl_idname, text='VRM using KITSUNETSUKI Asset Tools (.vrm)')



import logging
from collections import deque

import numpy as np
from panda3d.core import NodePath, Material

from metadrive.component.vehicle_navigation_module.base_navigation import BaseNavigation
from metadrive.engine.asset_loader import AssetLoader
from metadrive.utils.coordinates_shift import panda_vector
from metadrive.utils.math import norm, clip
from metadrive.utils.math import panda_vector
from metadrive.utils.math import wrap_to_pi

logger = logging.getLogger(__file__)


class TrajectoryNavigation(BaseNavigation):
    """
    This module enabling follow a given reference trajectory given a map
    """
    DISCRETE_LEN = 2  # m
    CHECK_POINT_INFO_DIM = 2
    NUM_WAY_POINT = 10
    NAVI_POINT_DIST = 30  # m, used to clip value, should be greater than DISCRETE_LEN * MAX_NUM_WAY_POINT

    def __init__(
        self,
        show_navi_mark: bool = False,
        random_navi_mark_color=False,
        show_dest_mark=False,
        show_line_to_dest=False,
        panda_color=None,
        name=None,
        vehicle_config=None
    ):
        if show_dest_mark or show_line_to_dest:
            logging.warning("show_dest_mark and show_line_to_dest are not supported in TrajectoryNavigation")
        super(TrajectoryNavigation, self).__init__(
            show_navi_mark=False,
            random_navi_mark_color=random_navi_mark_color,
            show_dest_mark=False,
            show_line_to_dest=False,
            panda_color=panda_color,
            name=name,
            vehicle_config=vehicle_config
        )
        self._route_completion = 0
        self.checkpoints = None  # All check points

        # for compatibility
        self.next_ref_lanes = None

        # override the show navi mark function here
        self._navi_point_model = None
        self._ckpt_vis_models = None
        if show_navi_mark and self._show_navi_info:
            self._ckpt_vis_models = [NodePath(str(i)) for i in range(self.NUM_WAY_POINT)]
            for model in self._ckpt_vis_models:
                if self._navi_point_model is None:
                    self._navi_point_model = AssetLoader.loader.loadModel(AssetLoader.file_path("models", "box.bam"))
                    self._navi_point_model.setScale(0.5)
                    if self.engine.use_render_pipeline:
                        material = Material()
                        material.setBaseColor((1, 1, 1, 1))
                        material.setShininess(128)
                        material.setEmission((1, 1, 1, 1))
                        self._navi_point_model.setMaterial(material, True)
                self._navi_point_model.instanceTo(model)
                model.reparentTo(self.origin)

        # should be updated every step after calling update_localization
        self.last_current_long = deque([0.0, 0.0], maxlen=2)
        self.last_current_lat = deque([0.0, 0.0], maxlen=2)
        self.last_current_heading_theta_at_long = deque([0.0, 0.0], maxlen=2)

    def reset(self, vehicle):
        super(TrajectoryNavigation, self).reset(current_lane=self.reference_trajectory)
        self.set_route()

    @property
    def reference_trajectory(self):
        return self.engine.map_manager.current_sdc_route

    @property
    def current_ref_lanes(self):
        return [self.reference_trajectory]

    def set_route(self):
        self.checkpoints = self.discretize_reference_trajectory()
        num_way_point = min(len(self.checkpoints), self.NUM_WAY_POINT)

        self._navi_info.fill(0.0)
        self.next_ref_lanes = None
        if self._dest_node_path is not None:
            check_point = self.reference_trajectory.end
            self._dest_node_path.setPos(panda_vector(check_point[0], check_point[1], 1))

    def discretize_reference_trajectory(self):
        ret = []
        length = self.reference_trajectory.length
        num = int(length / self.DISCRETE_LEN)
        for i in range(num):
            ret.append(self.reference_trajectory.position(i * self.DISCRETE_LEN, 0))
        ret.append(self.reference_trajectory.end)
        return ret

    def update_localization(self, ego_vehicle):
        """
        It is called every step
        """
        if self.reference_trajectory is None:
            return

        # Update ckpt index
        long, lat = self.reference_trajectory.local_coordinates(ego_vehicle.position)
        heading_theta_at_long = self.reference_trajectory.heading_theta_at(long)
        self.last_current_heading_theta_at_long.append(heading_theta_at_long)
        self.last_current_long.append(long)
        self.last_current_lat.append(lat)

        next_idx = max(int(long / self.DISCRETE_LEN) + 1, 0)
        next_idx = min(next_idx, len(self.checkpoints) - 1)
        end_idx = min(next_idx + self.NUM_WAY_POINT, len(self.checkpoints))
        ckpts = self.checkpoints[next_idx:end_idx]
        diff = self.NUM_WAY_POINT - len(ckpts)
        assert diff >= 0, "Number of Navigation points error!"
        if diff > 0:
            ckpts += [self.checkpoints[-1] for _ in range(diff)]

        # target_road_1 is the road segment the vehicle is driving on.
        self._navi_info.fill(0.0)
        for k, ckpt in enumerate(ckpts):
            start = k * self.CHECK_POINT_INFO_DIM
            end = (k + 1) * self.CHECK_POINT_INFO_DIM
            self._navi_info[start:end], lanes_heading = self._get_info_for_checkpoint(ckpt, ego_vehicle)
            if self._show_navi_info and self._ckpt_vis_models is not None:
                pos_of_goal = ckpt
                self._ckpt_vis_models[k].setPos(panda_vector(pos_of_goal[0], pos_of_goal[1], 1.8))
                self._ckpt_vis_models[k].setH(self._goal_node_path.getH() + 3)

        self._navi_info[end] = clip((lat / self.engine.global_config["max_lateral_dist"] + 1) / 2, 0.0, 1.0)
        self._navi_info[end + 1] = clip(
            (wrap_to_pi(heading_theta_at_long - ego_vehicle.heading_theta) / np.pi + 1) / 2, 0.0, 1.0
        )

        # Use RC as the only criterion to determine arrival in Scenario env.
        self._route_completion = long / self.reference_trajectory.length

    def get_current_lateral_range(self, current_position, engine) -> float:
        return self.current_lane.width * 2

    def get_current_lane_width(self) -> float:
        return self.current_lane.width

    def get_current_lane_num(self) -> float:
        return 1

    @classmethod
    def _get_info_for_checkpoint(cls, checkpoint, ego_vehicle):
        navi_information = []
        # Project the checkpoint position into the target vehicle's coordination, where
        # +x is the heading and +y is the right hand side.
        dir_vec = checkpoint - ego_vehicle.position  # get the vector from center of vehicle to checkpoint
        dir_norm = norm(dir_vec[0], dir_vec[1])
        if dir_norm > cls.NAVI_POINT_DIST:  # if the checkpoint is too far then crop the direction vector
            dir_vec = dir_vec / dir_norm * cls.NAVI_POINT_DIST
        ckpt_in_heading, ckpt_in_rhs = ego_vehicle.convert_to_local_coordinates(
            dir_vec, 0.0
        )  # project to ego vehicle's coordination

        # Dim 1: the relative position of the checkpoint in the target vehicle's heading direction.
        navi_information.append(clip((ckpt_in_heading / cls.NAVI_POINT_DIST + 1) / 2, 0.0, 1.0))

        # Dim 2: the relative position of the checkpoint in the target vehicle's right hand side direction.
        navi_information.append(clip((ckpt_in_rhs / cls.NAVI_POINT_DIST + 1) / 2, 0.0, 1.0))

        return navi_information

    def destroy(self):
        self.checkpoints = None
        # self.current_ref_lanes = None
        self.next_ref_lanes = None
        self.final_lane = None
        self._current_lane = None
        # self.reference_trajectory = None
        super(TrajectoryNavigation, self).destroy()

    def before_reset(self):
        self.checkpoints = None
        # self.current_ref_lanes = None
        self.next_ref_lanes = None
        self.final_lane = None
        self._current_lane = None
        # self.reference_trajectory = None

    @property
    def route_completion(self):
        return self._route_completion

    @classmethod
    def get_navigation_info_dim(cls):
        return cls.NUM_WAY_POINT * cls.CHECK_POINT_INFO_DIM + 2

    @property
    def last_longitude(self):
        return self.last_current_long[0]

    @property
    def current_longitude(self):
        return self.last_current_long[1]

    @property
    def last_lateral(self):
        return self.last_current_lat[0]

    @property
    def current_lateral(self):
        return self.last_current_lat[1]

    @property
    def last_heading_theta_at_long(self):
        return self.last_current_heading_theta_at_long[0]

    @property
    def current_heading_theta_at_long(self):
        return self.last_current_heading_theta_at_long[1]


WaymoTrajectoryNavigation = TrajectoryNavigation
NuPlanTrajectoryNavigation = TrajectoryNavigation



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from direct.stdpy.file import join

from metadrive.render_pipeline.rpcore.rpobject import RPObject


class BasePlugin(RPObject):
    """ This is the base class for all plugins. All plugin classes derive from
    this class. Additionally there are a lot of helpful functions provided,
    such as creating render stages. """

    # Plugins can set this to indicate they have some C++ code, so they cannot
    # be used with the python version.
    native_only = False

    # Plugins can set this to require other plugins
    required_plugins = tuple()

    def __init__(self, pipeline):
        """ Inits the plugin """
        self._pipeline = pipeline
        self._assigned_stages = []
        self.plugin_id = str(self.__class__.__module__).split(".")[-2]
        RPObject.__init__(self, "plugin:" + self.plugin_id)
        self._set_debug_color("magenta", "bright")

    @property
    def base_path(self):
        """ Returns the path to the root directory of the plugin """
        return "/$$rp/rpplugins/{}/".format(self.plugin_id)

    def get_resource(self, pth):
        """ Converts a local path from the plugins resource directory into
        an absolute path """
        return join(self.base_path, "resources", pth)

    def get_shader_resource(self, pth):
        """ Converts a local path from the plugins shader directory into
        an absolute path """
        return join(self.base_path, "shader", pth)

    def create_stage(self, stage_type):
        """ Shortcut to create a new render stage from a given class type. It
        also registers the stage to the stage manager, and links the stage
        to the current plugin instance. """
        stage_handle = stage_type(self._pipeline)
        self._pipeline.stage_mgr.add_stage(stage_handle)
        self._assigned_stages.append(stage_handle)
        return stage_handle

    def get_setting(self, setting_id, plugin_id=None):
        """ Returns the value of a setting given by its setting id. If plugin_id
        is set, returns the setting of the given plugin """
        return self._pipeline.plugin_mgr.settings[plugin_id or self.plugin_id][setting_id].value

    def get_daytime_setting(self, setting_id, plugin_id=None):
        """ Returns the value of a time of day setting given by its setting
        id. If plugin_id is set, returns the setting of a gien plugin """
        handle = self._pipeline.plugin_mgr.day_settings[plugin_id or self.plugin_id][setting_id]
        return handle.get_scaled_value_at(self._pipeline.daytime_mgr.time)

    def get_plugin_instance(self, plugin_id):
        """ Returns the instance of a different plugin, given by its id.
        This should be only used to access plugins which are also in the
        current plugins requirements """
        return self._pipeline.plugin_mgr.instances[plugin_id]

    def is_plugin_enabled(self, plugin_id):
        """ Returns whether a plugin is enabled and loaded, given is plugin id """
        return self._pipeline.plugin_mgr.is_plugin_enabled(plugin_id)

    def reload_shaders(self):
        """ Reloads all shaders of the plugin """
        for stage in self._assigned_stages:
            stage.reload_shaders()



import os.path
from distutils.dir_util import copy_tree

from metadrive.engine.asset_loader import AssetLoader
from metadrive.envs.scenario_env import ScenarioEnv
from metadrive.scenario.scenario_description import ScenarioDescription
from metadrive.scenario.utils import read_dataset_summary, read_scenario_data


def test_read_waymo_data():
    summary_dict, summary_list, mapping = read_dataset_summary(AssetLoader.file_path("waymo", return_raw_style=False))
    for p in summary_list:
        data = read_scenario_data(AssetLoader.file_path("waymo", mapping[p], p, return_raw_style=False))
        data.sanity_check(data, check_self_type=False, valid_check=False)
        print("Finish: ", p)


def test_read_data_no_summary():
    dir = AssetLoader.file_path("waymo", return_raw_style=False)
    new_dir = "test_read_copy_waymo"

    #  make fake dataset
    copy_tree(dir, new_dir)
    os.remove(os.path.join("test_read_copy_waymo", ScenarioDescription.DATASET.SUMMARY_FILE))
    env = ScenarioEnv(
        {
            "manual_control": True,
            "no_traffic": False,
            "start_scenario_index": 0,
            "show_coordinates": True,
            "num_scenarios": 3,
            "data_directory": new_dir,
        }
    )
    try:
        env.reset()
    finally:
        env.close()



"""
    (1) git clone git@github.com:argoai/argoverse-api.git
    (2) download argoverse maps at https://www.argoverse.org/data.html#download-link
    (3) uncompress the .tar file and rename the downloaded director to map_files
    (4) move the /map_files to argoverse-api repo
    (5) pip install -e /path_to_argoverse-api_directory/
"""
from argoverse.map_representation.map_api import ArgoverseMap

am = ArgoverseMap()



import logging

from panda3d.bullet import BulletWorld
from panda3d.core import Vec3

from metadrive.constants import CollisionGroup


class PhysicsWorld:
    def __init__(self, debug=False):
        # a dynamic world, moving objects or objects which react to other objects should be placed here
        self.dynamic_world = BulletWorld()
        CollisionGroup.set_collision_rule(self.dynamic_world)
        self.dynamic_world.setGravity(Vec3(0, 0, -9.81))  # set gravity
        # a static world which used to query position/overlap .etc. Don't implement doPhysics() in this world
        self.static_world = BulletWorld() if not debug else self.dynamic_world
        CollisionGroup.set_collision_rule(self.static_world)

    def report_bodies(self):
        dynamic_bodies = \
            self.dynamic_world.getNumRigidBodies() + self.dynamic_world.getNumGhosts() + self.dynamic_world.getNumVehicles()
        static_bodies = \
            self.static_world.getNumRigidBodies() + self.static_world.getNumGhosts() + self.static_world.getNumVehicles()
        return "dynamic bodies:{}, static_bodies: {}".format(dynamic_bodies, static_bodies)

    def destroy(self):
        self.dynamic_world.clearDebugNode()
        self.dynamic_world.clearContactAddedCallback()
        self.dynamic_world.clearFilterCallback()

        self.static_world.clearDebugNode()
        self.static_world.clearContactAddedCallback()
        self.static_world.clearFilterCallback()

        self.dynamic_world = None
        self.static_world = None

    def __del__(self):
        logging.debug("Physics world is destroyed successfully!")



from metadrive.component.map.base_map import BaseMap
from metadrive.component.map.pg_map import MapGenerateMethod
from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.utils import setup_logger

if __name__ == "__main__":
    setup_logger(True)
    env = MetaDriveEnv(
        {
            "num_scenarios": 1,
            "traffic_density": 0.1,
            "traffic_mode": "hybrid",
            "start_seed": 5,
            # "debug_physics_world": True,
            "pstats": True,
            # "controller":"joystick",
            "manual_control": True,
            "use_render": True,
            "decision_repeat": 5,
            "rgb_clip": True,
            "debug": True,
            "map_config": {
                BaseMap.GENERATE_TYPE: MapGenerateMethod.BIG_BLOCK_SEQUENCE,
                BaseMap.GENERATE_CONFIG: "yY",
                BaseMap.LANE_WIDTH: 3.5,
                BaseMap.LANE_NUM: 3,
            },
            "driving_reward": 1.0,
            "vehicle_config": {
                "show_lidar": False,
                "show_side_detector": True,
                "show_lane_line_detector": True,
                "lane_line_detector": {
                    "num_lasers": 100
                }
            }
        }
    )

    o, _ = env.reset()
    # print("vehicle num", len(env.engine.traffic_manager.vehicles))
    for i in range(1, 100000):
        o, r, tm, tc, info = env.step([0, 1])
        info["fuel"] = env.vehicle.energy_consumption
        env.render(
            text={
                "left": env.vehicle.dist_to_left_side,
                "right": env.vehicle.dist_to_right_side,
                "white_lane_line": env.vehicle.on_white_continuous_line
            }
        )
        if tm or tc:
            # print("Reset")
            env.reset()
    env.close()



import time

from metadrive.component.map.pg_map import PGMap
from metadrive.engine.engine_utils import initialize_engine, close_engine
from metadrive.envs import MetaDriveEnv

try:
    from reprlib import repr
except ImportError:
    pass

from metadrive.tests.test_functionality.test_memory_leak_engine import process_memory, total_size


def test_pg_map_memory_leak():

    try:
        default_config = MetaDriveEnv.default_config()

        # default_config["map_config"]["config"] = "SSS"
        # default_config["map_config"]["type"] = "block_sequence"

        default_config["map_config"]["config"] = 1

        # default_config["debug"] = True
        # default_config["debug_physics_world"] = True

        close_engine()
        engine = initialize_engine(default_config)

        ct = time.time()
        last_lm = cm = process_memory()
        last_mem = 0.0
        for t in range(50):
            lt = time.time()

            our_map = PGMap(default_config["map_config"])

            # our_map.blocks.clear()
            # our_map.blocks = []

            size = total_size(our_map)
            # print("map size: {:,}".format(size))
            del our_map

            # print(engine.physics_world.report_bodies())

            nlt = time.time()
            lm = process_memory()
            # print(
            #     "After {} Iters, Time {:.3f} Total Time {:.3f}, Memory Usage {:,} Memory Change {:,}".format(
            # t + 1, nlt - lt, nlt - ct, lm - cm, lm - last_lm
            #     )
            # )
            # last_lm = lm
            # if t > 100:
            #     assert abs((lm - cm) - last_mem) < 1024  # Memory should not have change > 1KB
            # last_mem = lm - cm

    finally:
        close_engine()

    assert lm - cm < 1024 * 1024 * 120, "We expect will cause 100M memory leak with 50 iterations."


if __name__ == "__main__":
    test_pg_map_memory_leak()



import time

from metadrive.constants import RENDER_MODE_ONSCREEN


class ForceFPS:
    UNLIMITED = "UnlimitedFPS"
    FORCED = "ForceFPS"

    def __init__(self, engine):
        # self.engine = engine
        if engine.global_config["force_render_fps"] is None:
            interval = engine.global_config["physics_world_step_size"]
        else:
            interval = 1 / engine.global_config["force_render_fps"]
        self.engine = engine
        fps = 1 / interval
        self.init_fps = fps
        if engine.mode == RENDER_MODE_ONSCREEN:
            self.state = self.FORCED
            self.engine.taskMgr.add(self.force_fps_task, "force_fps")
            self.fps = fps
        else:
            self.state = self.UNLIMITED
            self.fps = None

    @property
    def interval(self):
        return 1 / self.fps if self.fps is not None else None

    def tick(self):
        # print("Force fps, now: ", self.last)
        sim_interval = self.engine.task_manager.globalClock.getDt()
        if self.interval and sim_interval < self.interval:
            time.sleep(self.interval - sim_interval)

    def toggle(self):
        if self.state == self.UNLIMITED:
            self.engine.task_manager.add(self.force_fps_task, "force_fps")
            self.state = self.FORCED
            self.fps = self.init_fps
        elif self.state == self.FORCED:
            self.engine.task_manager.remove("force_fps")
            self.state = self.UNLIMITED
            self.fps = None

    def disable(self):
        if self.engine.task_manager.hasTaskNamed("force_fps"):
            self.engine.task_manager.remove("force_fps")
        self.state = self.UNLIMITED
        self.fps = None

    def enable(self):
        if not self.engine.task_manager.hasTaskNamed("force_fps"):
            self.engine.task_manager.add(self.force_fps_task, "force_fps")
        self.state = self.FORCED
        self.fps = self.init_fps

    def force_fps_task(self, task):
        self.tick()
        return task.cont

    @property
    def real_time_simulation(self):
        return self.state == self.FORCED and self.engine.mode == RENDER_MODE_ONSCREEN



from metadrive.base_class.base_object import BaseObject
from metadrive.constants import CamMask
from metadrive.scenario.scenario_description import ScenarioDescription
from metadrive.constants import MetaDriveType
from metadrive.engine.asset_loader import AssetLoader
from metadrive.utils.pg.utils import generate_static_box_physics_body


class BaseTrafficLight(BaseObject):
    """
    Traffic light should be associated with a lane before using. It is basically an unseen wall object on the route, so
    actors have to react to it.
    """
    AIR_WALL_LENGTH = 0.25
    AIR_WALL_HEIGHT = 1.5
    TRAFFIC_LIGHT_HEIGHT = 3
    TRAFFIC_LIGHT_MODEL = {}
    LIGHT_VIS_HEIGHT = 0.8
    LIGHT_VIS_WIDTH = 0.8
    PLACE_LONGITUDE = 5

    def __init__(
        self, lane, position=None, name=None, random_seed=None, config=None, escape_random_seed_assertion=False
    ):
        super(BaseTrafficLight, self).__init__(name, random_seed, config, escape_random_seed_assertion)
        self.lane = lane
        self.status = MetaDriveType.LIGHT_UNKNOWN

        self.lane_width = lane.width_at(0) if lane else 4
        air_wall = generate_static_box_physics_body(
            self.AIR_WALL_LENGTH,
            self.lane_width,
            self.AIR_WALL_HEIGHT,
            object_id=self.id,
            type_name=MetaDriveType.TRAFFIC_LIGHT,
            ghost_node=True,
        )
        self.add_body(air_wall, add_to_static_world=True)

        if position is None:
            # auto determining
            position = lane.position(self.PLACE_LONGITUDE, 0)

        self.set_position(position, self.AIR_WALL_HEIGHT / 2)
        self.set_heading_theta(lane.heading_theta_at(self.PLACE_LONGITUDE) if lane else 0)
        self.current_light = None

        if self.render:
            if len(BaseTrafficLight.TRAFFIC_LIGHT_MODEL) == 0:
                for color in ["green", "red", "yellow", "unknown"]:
                    model = self.loader.loadModel(
                        AssetLoader.file_path("models", "traffic_light", "{}.gltf".format(color))
                    )
                    model.setPos(0, 0, self.TRAFFIC_LIGHT_HEIGHT)
                    model.setH(-90)
                    model.hide(CamMask.Shadow)
                    BaseTrafficLight.TRAFFIC_LIGHT_MODEL[color] = model
            self.origin.setScale(0.5, 1.2, 1.2)

    def set_status(self, status):
        """
        People should overwrite this method to parse traffic light status and to determine which traffic light to set
        """
        pass

    def set_green(self):
        if self.render:
            if self.current_light is not None:
                self.current_light.detachNode()
            self.current_light = BaseTrafficLight.TRAFFIC_LIGHT_MODEL["green"].instanceTo(self.origin)
        self.status = MetaDriveType.LIGHT_GREEN

    def set_red(self):
        if self.render:
            if self.current_light is not None:
                self.current_light.detachNode()
            self.current_light = BaseTrafficLight.TRAFFIC_LIGHT_MODEL["red"].instanceTo(self.origin)
        self.status = MetaDriveType.LIGHT_RED

    def set_yellow(self):
        if self.render:
            if self.current_light is not None:
                self.current_light.detachNode()
            self.current_light = BaseTrafficLight.TRAFFIC_LIGHT_MODEL["yellow"].instanceTo(self.origin)
        self.status = MetaDriveType.LIGHT_YELLOW

    def set_unknown(self):
        if self.render:
            if self.current_light is not None:
                self.current_light.detachNode()
            self.current_light = BaseTrafficLight.TRAFFIC_LIGHT_MODEL["unknown"].instanceTo(self.origin)
        self.status = MetaDriveType.LIGHT_UNKNOWN

    def destroy(self):
        super(BaseTrafficLight, self).destroy()
        self.lane = None

    @property
    def top_down_color(self):
        status = self.status
        if status == MetaDriveType.LIGHT_GREEN:
            return [0, 255, 0]
        if status == MetaDriveType.LIGHT_RED:
            return [255, 0, 0]
        if status == MetaDriveType.LIGHT_YELLOW:
            return [255, 255, 0]
        if status == MetaDriveType.LIGHT_UNKNOWN:
            return [180, 180, 180]

    @property
    def top_down_width(self):
        return 1.5

    @property
    def top_down_length(self):
        return 1.5

    def set_action(self, *args, **kwargs):
        return self.set_status(*args, **kwargs)

    def get_state(self):
        pos = self.position
        state = {
            ScenarioDescription.TRAFFIC_LIGHT_POSITION: pos,
            ScenarioDescription.TRAFFIC_LIGHT_STATUS: self.status,
            ScenarioDescription.TYPE: type(self)
        }
        return state

    @property
    def LENGTH(self):
        return self.AIR_WALL_LENGTH

    @property
    def WIDTH(self):
        return self.lane_width



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from __future__ import print_function, division
from metadrive.render_pipeline.rplibs.six.moves import range  # pylint: disable=import-error

from panda3d.core import LVecBase4i, LVecBase4


class ShadowAtlas(object):
    """ Please refer to the native C++ implementation for docstrings and comments.
    This is just the python implementation, which does not contain documentation! """
    def __init__(self, size, tile_size=32):
        self._size = size
        self._tile_size = tile_size
        self._num_used_tiles = 0
        self.init_tiles()

    def init_tiles(self):
        self._num_tiles = self._size // self._tile_size

        def row():
            return [False for i in range(self._num_tiles)]  # pylint: disable=unused-variable

        self._flags = [row() for j in range(self._num_tiles)]  # pylint: disable=unused-variable

    def get_num_used_tiles(self):
        return self._num_used_tiles

    num_used_tiles = property(get_num_used_tiles)

    def get_coverage(self):
        return self._num_used_tiles / float(self._num_tiles**2)

    coverage = property(get_coverage)

    def reserve_region(self, x, y, w, h):
        self._num_used_tiles += w * h
        for x_offset in range(w):
            for y_offset in range(h):
                self._flags[x + x_offset][y + y_offset] = True

    def find_and_reserve_region(self, tile_width, tile_height):
        for x in range(self._num_tiles - tile_height + 1):
            for y in range(self._num_tiles - tile_width + 1):
                if self.region_is_free(x, y, tile_width, tile_height):
                    self.reserve_region(x, y, tile_width, tile_height)
                    return LVecBase4i(x, y, tile_width, tile_height)
        print("Failed to find a free region of size", tile_width, "x", tile_height)
        return LVecBase4i(-1)

    def free_region(self, region):
        self._num_used_tiles -= region.z * region.w
        for x in range(region.z):
            for y in range(region.w):
                self._flags[region.x + x][region.y + y] = False

    def get_tile_size(self):
        return self._tile_size

    def region_is_free(self, x, y, w, h):
        for x_offset in range(w):
            for y_offset in range(h):
                if self._flags[x + x_offset][y + y_offset]:
                    return False
        return True

    def get_required_tiles(self, resolution):
        if resolution % self._tile_size != 0:
            print("ShadowAtlas: Invalid atlas resolution!")
            return
        return resolution // self._tile_size

    def region_to_uv(self, region):
        flt = LVecBase4(region.x, region.y, region.z, region.w)
        return flt * (self._tile_size / self._size)



import argparse
import logging
import time

import numpy as np

from metadrive import MultiAgentRoundaboutEnv
from metadrive.utils import setup_logger

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("--num-steps", "-n", default=10_000, type=int, help="Total steps of profiling.")
    args = parser.parse_args()

    print("Start to profile the efficiency of MetaDrive Multi-agent Roundabout environment!")
    setup_logger(debug=False)
    env = MultiAgentRoundaboutEnv(dict(start_seed=1010, log_level=logging.FATAL))
    obs, _ = env.reset()
    start = time.time()
    reset_used_time = 0
    action = [0.0, 1.]
    total_steps = args.num_steps
    vehicle_num = [len(env.vehicles)]
    for s in range(total_steps):
        o, r, tm, tc, i = env.step({k: action for k in env.vehicles})
        if tm["__all__"]:
            start_reset = time.time()
            env.reset()
            vehicle_num.append(len(env.vehicles))
            reset_used_time += time.time() - start_reset
        if (s + 1) % 100 == 0:
            print(
                "Finish {}/{} simulation steps. Time elapse: {:.4f}. Average FPS: {:.4f}, Average number of "
                "vehicles: {:.4f}".format(
                    s + 1, total_steps,
                    time.time() - start, (s + 1) / (time.time() - start - reset_used_time), np.mean(vehicle_num)
                )
            )
    print(
        "Total Time Elapse: {:.3f}, average FPS: {:.3f}, average number of vehicles: {:.3f}.".format(
            time.time() - start, total_steps / (time.time() - start - reset_used_time), np.mean(vehicle_num)
        )
    )



from metadrive.component.pgblock.curve import Curve
from metadrive.component.pgblock.first_block import FirstPGBlock
from metadrive.component.pgblock.std_t_intersection import StdTInterSection
from metadrive.component.pgblock.straight import Straight
from metadrive.component.road_network.node_road_network import NodeRoadNetwork
from metadrive.tests.vis_block.vis_block_base import TestBlock

if __name__ == "__main__":
    test = TestBlock(True)
    from metadrive.engine.asset_loader import initialize_asset_loader

    initialize_asset_loader(test)

    global_network = NodeRoadNetwork()
    first = FirstPGBlock(global_network, 3.0, 2, test.render, test.world, 1)

    curve = Curve(1, first.get_socket(0), global_network, 1)
    curve.construct_block(test.render, test.world)

    straight = Straight(2, curve.get_socket(0), global_network, 1)
    straight.construct_block(test.render, test.world)

    intersection = StdTInterSection(3, straight.get_socket(0), global_network, 1)
    # print(intersection.construct_block(test.render, test.world))
    id = 4
    for socket_idx in range(intersection.SOCKET_NUM):
        block = Curve(id, intersection.get_socket(socket_idx), global_network, id + 1)
        block.construct_block(test.render, test.world)
        id += 1
    test.show_bounding_box(global_network)
    test.run()



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from __future__ import division
from functools import partial

from panda3d.core import Vec3

from metadrive.render_pipeline.rplibs.yaml import load_yaml_file

from metadrive.render_pipeline.rpcore.native import NATIVE_CXX_LOADED
from metadrive.render_pipeline.rpcore.gui.draggable_window import DraggableWindow
from metadrive.render_pipeline.rpcore.gui.labeled_checkbox import LabeledCheckbox
from metadrive.render_pipeline.rpcore.gui.checkbox_collection import CheckboxCollection


class RenderModeSelector(DraggableWindow):
    """ Window which offers the user to select a render mode to apply """
    def __init__(self, pipeline, parent):
        DraggableWindow.__init__(self, width=690, height=340, parent=parent, title="Select render mode")
        self._pipeline = pipeline
        self._selected_mode = ""
        self._create_components()
        self.hide()

    def _create_components(self):
        """ Internal method to init the components """
        DraggableWindow._create_components(self)
        self._content_node = self._node.attach_new_node("content")
        self._populate_content()

    def _populate_content(self):
        """ Populates the windows content """
        self._content_node.node().remove_all_children()

        # Reload config each time the window is opened so its easy to add new
        # render modes
        config = load_yaml_file("/$$rpconfig/debugging.yaml")

        debugger_content = self._content_node.attach_new_node("RenderModes")
        debugger_content.set_z(-20)
        debugger_content.set_x(20)

        render_modes = [("Default", "", False, "", False)]

        # Read modes from configuration
        for mode in config["render_modes"]:
            data = [mode["name"], mode["key"]]
            data.append(mode.get("cxx_only", False))
            data.append(mode.get("requires", ""))
            data.append(mode.get("special", False))
            render_modes.append(data)

        collection = CheckboxCollection()

        max_column_height = 9

        for idx, (mode, mode_id, requires_cxx, requires_plugin, special) in enumerate(render_modes):
            offs_y = (idx % max_column_height) * 24 + 35
            offs_x = (idx // max_column_height) * 220
            enabled = True
            if requires_cxx and not NATIVE_CXX_LOADED:
                enabled = False

            if requires_plugin:
                if not self._pipeline.plugin_mgr.is_plugin_enabled(requires_plugin):
                    enabled = False

            box = LabeledCheckbox(
                parent=debugger_content,
                x=offs_x,
                y=offs_y,
                text=mode.upper(),
                text_color=Vec3(0.4),
                radio=True,
                chb_checked=(mode_id == self._selected_mode),
                chb_callback=partial(self._set_render_mode, mode_id, special),
                text_size=14,
                expand_width=230,
                enabled=enabled
            )
            collection.add(box.checkbox)

    def _set_render_mode(self, mode_id, special, value):
        """ Callback which gets called when a render mode got selected """
        if not value:
            return

        to_remove = []
        for define in self._pipeline.stage_mgr.defines:
            if define.startswith("_RM_"):
                to_remove.append(define)
        for define in to_remove:
            del self._pipeline.stage_mgr.defines[define]

        if mode_id == "":
            self._pipeline.stage_mgr.defines["ANY_DEBUG_MODE"] = 0
        else:
            # Don't activate the generic debugging mode for special modes. This
            # is for modes like luminance which expect the scene to be rendered
            # unaltered.
            self._pipeline.stage_mgr.defines["ANY_DEBUG_MODE"] = 0 if special else 1
            self._pipeline.stage_mgr.defines["_RM_" + mode_id] = 1

        self._selected_mode = mode_id
        self._pipeline.reload_shaders()

    def toggle(self):
        """ Toggles the visibility of this windows """
        if self._visible:
            self.hide()
        else:
            self._populate_content()
            self.show()
